"""
Likelihood Models
-----------------

The likelihood models contain all the logic needed to train and use Darts' neural network models in
a probabilistic way. This essentially means computing an appropriate training loss and sample from the
distribution, given the parameters of the distribution.

By default, all versions will be trained using their negative log likelihood as a loss function
(hence performing maximum likelihood estimation when training the model).
However, most likelihoods also optionally support specifying time-independent "prior"
beliefs about the distribution parameters.
In such cases, the a KL-divergence term is added to the loss in order to regularise it in the
direction of the specified prior distribution. (Note that this is technically not purely
a Bayesian approach as the priors are actual parameters values, and not distributions).
The parameter `prior_strength` controls the strength of the "prior" regularisation on the loss.

Some distributions (such as ``GaussianLikelihood``, and ``PoissonLikelihood``) are univariate,
in which case they are applied to model each component of multivariate series independently.
Some other distributions (such as ``DirichletLikelihood``) are multivariate,
in which case they will model all components of multivariate time series jointly.

Univariate likelihoods accept either scalar or array-like values for the optional prior parameters.
If a scalar is provided, it is used as a prior for all components of the series. If an array-like is provided,
the i-th value will be used as a prior for the i-th component of the series. Multivariate likelihoods
require array-like objects when specifying priors.

The target series used for training must always lie within the distribution's support, otherwise
errors will be raised during training. You can refer to the individual likelihoods' documentation
to see what is the support. Similarly, the prior parameters also have to lie in some pre-defined domains.
"""

# TODO: Table on README listing distribution, possible priors and wiki article
from darts.utils.utils import raise_if_not

from abc import ABC, abstractmethod
from typing import Optional, Tuple
import collections

import numpy as np
import torch
import torch.nn as nn
from torch.distributions.kl import kl_divergence
from torch.distributions import (Normal as _Normal,
                                 Poisson as _Poisson,
                                 NegativeBinomial as _NegativeBinomial,
                                 Bernoulli as _Bernoulli,
                                 Gamma as _Gamma,
                                 Gumbel as _Gumbel,
                                 Laplace as _Laplace,
                                 Beta as _Beta,
                                 Exponential as _Exponential,
                                 MultivariateNormal as _MultivariateNormal,
                                 Dirichlet as _Dirichlet,
                                 Geometric as _Geometric,
                                 Cauchy as _Cauchy,
                                 ContinuousBernoulli as _ContinuousBernoulli,
                                 HalfNormal as _HalfNormal,
                                 LogNormal as _LogNormal,
                                 Weibull as _Weibull
                                 )

MIN_CAUCHY_GAMMA_SAMPLING = 1e-100

# Some utils for checking parameters' domains
def _check(param, predicate, param_name, condition_str):
    if param is None: return
    if isinstance(param, (collections.Sequence, np.ndarray)):
        raise_if_not(all(predicate(p) for p in param),
                     'All provided parameters {} must be {}.'.format(param_name, condition_str))
    else:
        raise_if_not(predicate(param), 'The parameter {} must be {}.'.format(param_name, condition_str))


def _check_strict_positive(param, param_name=''):
    _check(param, lambda p: p > 0, param_name, 'strictly positive')


def _check_in_open_0_1_intvl(param, param_name=''):
    _check(param, lambda p: 0 < p < 1, param_name, 'in the open interval (0, 1)')


class Likelihood(ABC):
    def __init__(self, prior_strength=1.):
        """
        Abstract class for a likelihood model.
        """
        self.prior_strength = prior_strength

    def compute_loss(self, model_output: torch.Tensor, target: torch.Tensor):
        """
        Computes a loss from a `model_output`, which represents the parameters of a given probability
        distribution for every ground truth value in `target`, and the `target` itself.
        """
        params_out = self._params_from_output(model_output)
        loss = self._nllloss(params_out, target)

        prior_params = self._prior_params
        use_prior = prior_params is not None and any(p is not None for p in prior_params)
        if use_prior:
            out_distr = self._distr_from_params(params_out)
            device = params_out[0].device
            prior_params = tuple([
                # use model output as "prior" for parameters not specified as prior
                torch.tensor(prior_params[i]).to(device) if prior_params[i] is not None else params_out[i]
                for i in range(len(prior_params))
            ])
            prior_distr = self._distr_from_params(prior_params)

            # Loss regularization using the prior distribution
            loss += self.prior_strength * torch.mean(kl_divergence(prior_distr, out_distr))

        return loss

    def _nllloss(self, params_out, target):
        """
        This is the basic way to compute the NLL loss. It can be overwritten by likelihoods for which
        PyTorch proposes a numerically better NLL loss.
        """
        out_distr = self._distr_from_params(params_out)
        return -out_distr.log_prob(target).mean()

    @property
    def _prior_params(self):
        """
        Has to be overwritten by the Likelihood objects supporting specifying a prior distribution on the
        outputs. If it returns None, no prior will be used and the model will be trained with plain maximum likelihood.
        """
        return None

    @abstractmethod
    def _distr_from_params(self, params: Tuple) -> torch.distributions.Distribution:
        """
        Returns a torch distribution built with the specified params
        """
        pass

    @abstractmethod
    def _params_from_output(self, model_output: torch.Tensor):
        """
        Returns the distribution parameters, obtained from the raw model outputs
        (e.g. applies softplus or sigmoids to get parameters in the expected domains).
        """
        pass

    @abstractmethod
    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        """
        Samples a prediction from the probability distributions defined by the specific likelihood model
        and the parameters given in `model_output`.
        """
        pass

    @property
    @abstractmethod
    def num_parameters(self) -> int:
        """
        Returns the number of parameters that define the probability distribution for one single
        target value.
        """
        pass


class GaussianLikelihood(Likelihood):
    def __init__(self, prior_mu=None, prior_sigma=None, prior_strength=1.):
        """
        Univariate Gaussian distribution.

        https://en.wikipedia.org/wiki/Normal_distribution

        - Univariate continuous distribution.
        - Support: :math:`\mathbb{R}`.
        - Parameters: mean :math:`\\mu \in \mathbb{R}`, standard deviation :math:`\\sigma > 0`.

        Parameters
        ----------
        prior_mu
            mean of the prior Gaussian distribution (default: None).
        prior_sigma
            standard deviation (or scale) of the prior Gaussian distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_mu = prior_mu
        self.prior_sigma = prior_sigma
        _check_strict_positive(self.prior_sigma, 'sigma')

        self.nllloss = nn.GaussianNLLLoss(reduction='mean', full=True)
        self.softplus = nn.Softplus()

        super().__init__(prior_strength)

    def _nllloss(self, params_out, target):
        means_out, sigmas_out = params_out
        return self.nllloss(means_out.contiguous(), target.contiguous(), sigmas_out.contiguous())

    @property
    def _prior_params(self):
        return self.prior_mu, self.prior_sigma

    def _distr_from_params(self, params):
        mu, sigma = params
        return _Normal(mu, sigma)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        mu, sigma = self._params_from_output(model_output)
        return torch.normal(mu, sigma)

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output):
        output_size = model_output.shape[-1]
        mu = model_output[:, :, :output_size // 2]
        sigma = self.softplus(model_output[:, :, output_size // 2:])
        return mu, sigma


class PoissonLikelihood(Likelihood):
    def __init__(self, prior_lambda=None, prior_strength=1.):
        """
        Poisson distribution. Can typically be used to model event counts during time intervals, when the events
        happen independently of the time since the last event.

        https://en.wikipedia.org/wiki/Poisson_distribution

        - Univariate discrete distribution
        - Support: :math:`\mathbb{N}_0` (natural numbers including 0).
        - Parameter: rate :math:`\\lambda > 0`.

        Parameters
        ----------
        prior_lambda
            rate :math:`\\lambda` of the prior Poisson distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_lambda = prior_lambda
        _check_strict_positive(self.prior_lambda, 'lambda')

        self.nllloss = nn.PoissonNLLLoss(log_input=False, full=True)
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    def _nllloss(self, params_out, target):
        lambda_out = params_out
        return self.nllloss(lambda_out, target)

    @property
    def _prior_params(self):
        return self.prior_lambda,

    def _distr_from_params(self, params):
        lmbda = params[0]
        return _Poisson(lmbda)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        model_lambda = self._params_from_output(model_output)
        return torch.poisson(model_lambda)

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output):
        lmbda = self.softplus(model_output)
        return lmbda


class NegativeBinomialLikelihood(Likelihood):
    def __init__(self):
        """
        Negative Binomial distribution.

        https://en.wikipedia.org/wiki/Negative_binomial_distribution

        It does not support priors.

        - Univariate discrete distribution.
        - Support: :math:`\mathbb{N}_0` (natural numbers including 0).
        - Parameters: number of failures :math:`r > 0`, success probability :math:`p \in (0, 1)`.

        Behind the scenes the distribution is reparameterized so that the actual outputs of the
        network are in terms of the mean :math:`\\mu` and shape :math:`\\alpha`.
        """
        self.softplus = nn.Softplus()
        super().__init__()

    @property
    def _prior_params(self):
        return None

    @staticmethod
    def _get_r_and_p_from_mu_and_alpha(mu, alpha):
        # See https://en.wikipedia.org/wiki/Negative_binomial_distribution for the different parametrizations
        r = 1. / alpha
        p = r / (mu + r)
        return r, p

    def _distr_from_params(self, params):
        mu, alpha = params
        r, p = NegativeBinomialLikelihood._get_r_and_p_from_mu_and_alpha(mu, alpha)
        return _NegativeBinomial(r, p)

    def sample(self, model_output: torch.Tensor):
        mu, alpha = self._params_from_output(model_output)
        r, p = NegativeBinomialLikelihood._get_r_and_p_from_mu_and_alpha(mu, alpha)
        distr = _NegativeBinomial(r, p)
        return distr.sample()

    def _params_from_output(self, model_output):
        output_size = model_output.shape[-1]
        mu = self.softplus(model_output[:, :, :output_size // 2])
        alpha = self.softplus(model_output[:, :, output_size // 2:])
        return mu, alpha

    @property
    def num_parameters(self) -> int:
        return 2


class BernoulliLikelihood(Likelihood):
    def __init__(self, prior_p=None, prior_strength=1.):
        """
        Bernoulli distribution.

        https://en.wikipedia.org/wiki/Bernoulli_distribution

        - Univariate discrete distribution.
        - Support: :math:`\{0, 1\}`.
        - Parameter: probability :math:`p \in (0, 1)`.

        Parameters
        ----------
        prior_p
            probability :math:`p` of the prior Bernoulli distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_p = prior_p
        _check_in_open_0_1_intvl(self.prior_p, 'p')

        self.sigmoid = nn.Sigmoid()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_p,

    def _distr_from_params(self, params):
        p = params[0]
        return _Bernoulli(p)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        model_p = self._params_from_output(model_output)
        return torch.bernoulli(model_p)

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output: torch.Tensor):
        p = self.sigmoid(model_output)
        return p


class BetaLikelihood(Likelihood):
    def __init__(self, prior_alpha=None, prior_beta=None, prior_strength=1.):
        """
        Beta distribution.

        https://en.wikipedia.org/wiki/Beta_distribution

        - Univariate continuous distribution.
        - Support: open interval :math:`(0,1)`
        - Parameters: shape parameters :math:`\\alpha > 0` and :math:`\\beta > 0`.

        Parameters
        ----------
        prior_alpha
            shape parameter :math:`\\alpha` of the Beta distribution, strictly positive (default: None)
        prior_beta
            shape parameter :math:`\\beta` distribution, strictly positive (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_alpha = prior_alpha
        self.prior_beta = prior_beta
        _check_strict_positive(self.prior_alpha, 'alpha')
        _check_strict_positive(self.prior_beta, 'beta')

        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_alpha, self.prior_beta

    def _distr_from_params(self, params):
        alpha, beta = params
        return _Beta(alpha, beta)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        alpha, beta = self._params_from_output(model_output)
        distr = _Beta(alpha, beta)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output):
        output_size = model_output.shape[-1]
        alpha = self.softplus(model_output[:, :, :output_size // 2])
        beta = self.softplus(model_output[:, :, output_size // 2:])
        return alpha, beta


class CauchyLikelihood(Likelihood):
    def __init__(self, prior_xzero=None, prior_gamma=None, prior_strength=1.):
        """
        Cauchy Distribution.

        https://en.wikipedia.org/wiki/Cauchy_distribution

        - Univariate continuous distribution.
        - Support: :math:`\mathbb{R}`.
        - Parameters: location :math:`x_0 \in \mathbb{R}`, scale :math:`\gamma > 0`.

        Due to its fat tails, this distribution is typically harder to estimate,
        and your mileage may vary. Also be aware that it typically
        requires a large value for `num_samples` for sampling predictions.

        Parameters
        ----------
        prior_xzero
            location parameter :math:`x_0` of the Cauchy distribution (default: None)
        prior_gamma
            scale parameter :math:`\\gamma` of the Cauchy distribution, strictly positive (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_xzero = prior_xzero
        self.prior_gamma = prior_gamma
        _check_strict_positive(self.prior_gamma, 'gamma')

        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_xzero, self.prior_gamma

    def _distr_from_params(self, params):
        xzero, gamma = params
        return _Cauchy(xzero, gamma)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        xzero, gamma = self._params_from_output(model_output)

        # We need this hack as sometimes the output of the softplus is 0 in practice for Cauchy...
        gamma[gamma < MIN_CAUCHY_GAMMA_SAMPLING] = MIN_CAUCHY_GAMMA_SAMPLING

        distr = _Cauchy(xzero, gamma)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output):
        output_size = model_output.shape[-1]
        xzero = model_output[:, :, :output_size // 2]
        gamma = self.softplus(model_output[:, :, output_size // 2:])
        return xzero, gamma


class ContinuousBernoulliLikelihood(Likelihood):
    def __init__(self, prior_lambda=None, prior_strength=1.):
        """
        Continuous Bernoulli distribution.

        https://en.wikipedia.org/wiki/Continuous_Bernoulli_distribution

        - Univariate continuous distribution.
        - Support: open interval :math:`(0, 1)`.
        - Parameter: shape :math:`\\lambda \in (0,1)`

        Parameters
        ----------
        prior_lambda
            shape :math:`\\lambda` of the prior Continuous Bernoulli distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_lambda = prior_lambda
        _check_in_open_0_1_intvl(self.prior_lambda, 'lambda')

        self.sigmoid = nn.Sigmoid()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_lambda,

    def _distr_from_params(self, params):
        lmbda = params[0]
        return _ContinuousBernoulli(lmbda)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        model_lmbda = self._params_from_output(model_output)
        distr = _ContinuousBernoulli(model_lmbda)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output: torch.Tensor):
        lmbda = self.sigmoid(model_output)
        return lmbda


class DirichletLikelihood(Likelihood):
    def __init__(self, prior_alphas=None, prior_strength=1.):
        """
        Dirichlet distribution.

        https://en.wikipedia.org/wiki/Dirichlet_distribution

        - Multivariate continuous distribution, modeling all components of a time series jointly.
        - Support: The :math:`K`-dimensional simplex for series of dimension :math:`K`, i.e.,
          :math:`x_1, ..., x_K \\text{ with } x_i \in (0,1),\\; \\sum_i^K{x_i}=1`.
        - Parameter: concentrations :math:`\\alpha_1, ..., \\alpha_K` with :math:`\\alpha_i > 0`.

        Parameters
        ----------
        prior_alphas
            concentrations parameters :math:`\\alpha` of the prior Dirichlet distribution.
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_alphas = prior_alphas
        _check_strict_positive(self.prior_alphas)
        self.softmax = nn.Softmax(dim=2)
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_alphas,

    def _distr_from_params(self, params: Tuple):
        alphas = params[0]
        return _Dirichlet(alphas)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        alphas = self._params_from_output(model_output)
        distr = _Dirichlet(alphas)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 1  # 1 parameter per component

    def _params_from_output(self, model_output):
        alphas = self.softmax(model_output)  # take softmax over components
        return alphas


class ExponentialLikelihood(Likelihood):
    def __init__(self, prior_lambda=None, prior_strength=1.):
        """
        Exponential distribution.

        https://en.wikipedia.org/wiki/Exponential_distribution

        - Univariate continuous distribution.
        - Support: :math:`\mathbb{R}_{>0}`.
        - Parameter: rate :math:`\\lambda > 0`.

        Parameters
        ----------
        prior_lambda
            rate :math:`\\lambda` of the prior exponential distribution (default: None).
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_lambda = prior_lambda
        _check_strict_positive(self.prior_lambda, 'lambda')
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_lambda,

    def _distr_from_params(self, params: Tuple):
        lmbda = params[0]
        return _Exponential(lmbda)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        lmbda = self._params_from_output(model_output)
        distr = _Exponential(lmbda)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output: torch.Tensor):
        lmbda = self.softplus(model_output)
        return lmbda


class GammaLikelihood(Likelihood):
    def __init__(self, prior_alpha=None, prior_beta=None, prior_strength=1.):
        """
        Gamma distribution.

        https://en.wikipedia.org/wiki/Gamma_distribution

        - Univariate continuous distribution
        - Support: :math:`\mathbb{R}_{>0}`.
        - Parameters: shape :math:`\\alpha > 0` and rate :math:`\\beta > 0`.

        Parameters
        ----------
        prior_alpha
            shape :math:`\\alpha` of the prior gamma distribution (default: None).
        prior_beta
            rate :math:`\\beta` of the prior gamma distribution (default: None).
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_alpha = prior_alpha
        self.prior_beta = prior_beta
        _check_strict_positive(self.prior_alpha, 'alpha')
        _check_strict_positive(self.prior_beta, 'beta')
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_alpha, self.prior_beta

    def _distr_from_params(self, params: Tuple):
        alpha, beta = params
        return _Gamma(alpha, beta)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        alpha, beta = self._params_from_output(model_output)
        distr = _Gamma(alpha, beta)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output: torch.Tensor):
        output_size = model_output.shape[-1]
        alpha = self.softplus(model_output[:, :, :output_size // 2])
        beta = self.softplus(model_output[:, :, output_size // 2:])
        return alpha, beta


class GeometricLikelihood(Likelihood):
    def __init__(self, prior_p=None, prior_strength=1.):
        """
        Geometric distribution.

        https://en.wikipedia.org/wiki/Geometric_distribution

        - Univariate discrete distribution
        - Support: :math:`\mathbb{N}_0` (natural numbers including 0).
        - Parameter: success probability :math:`p \in (0, 1)`.

        Parameters
        ----------
        prior_p
            success probability :math:`p` of the prior geometric distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_p = prior_p
        _check_in_open_0_1_intvl(self.prior_p, 'p')
        self.sigmoid = nn.Sigmoid()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_p,

    def _distr_from_params(self, params: Tuple):
        p = params[0]
        return _Geometric(p)

    def sample(self, model_output):
        p = self._params_from_output(model_output)
        distr = _Geometric(p)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output: torch.Tensor):
        p = self.sigmoid(model_output)
        return p


class GumbelLikelihood(Likelihood):
    def __init__(self, prior_mu=None, prior_beta=None, prior_strength=1.):
        """
        Gumbel distribution.

        https://en.wikipedia.org/wiki/Gumbel_distribution

        - Univariate continuous distribution
        - Support: :math:`\mathbb{R}`.
        - Parameters: location :math:`\\mu \in \mathbb{R}` and scale :math:`\\beta > 0`.

        Parameters
        ----------
        prior_mu
            location :math:`\\mu` of the prior Gumbel distribution (default: None).
        prior_beta
            scale :math:`\\beta` of the prior Gumbel distribution (default: None).
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_mu = prior_mu
        self.prior_beta = prior_beta
        _check_strict_positive(self.prior_beta)
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_mu, self.prior_beta

    def _distr_from_params(self, params: Tuple):
        mu, beta = params
        return _Gumbel(mu, beta)

    def sample(self, model_output):
        mu, beta = self._params_from_output(model_output)
        distr = _Gumbel(mu, beta)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output: torch.Tensor):
        output_size = model_output.shape[-1]
        mu = model_output[:, :, :output_size // 2]
        beta = self.softplus(model_output[:, :, output_size // 2:])
        return mu, beta


class HalfNormalLikelihood(Likelihood):
    def __init__(self, prior_sigma=None, prior_strength=1.):
        """
        Half-normal distribution.

        https://en.wikipedia.org/wiki/Half-normal_distribution

        - Univariate continuous distribution.
        - Support: :math:`\mathbb{R}_{>0}`.
        - Parameter: rate :math:`\\sigma > 0`.

        Parameters
        ----------
        prior_sigma
            standard deviation :math:`\\sigma` of the prior half-normal distribution (default: None).
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_sigma = prior_sigma
        _check_strict_positive(self.prior_sigma, 'sigma')
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_sigma,

    def _distr_from_params(self, params: Tuple):
        sigma = params[0]
        return _HalfNormal(sigma)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        sigma = self._params_from_output(model_output)
        distr = _HalfNormal(sigma)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 1

    def _params_from_output(self, model_output: torch.Tensor):
        sigma = self.softplus(model_output)
        return sigma


class LaplaceLikelihood(Likelihood):
    def __init__(self, prior_mu=None, prior_b=None, prior_strength=1.):
        """
        Laplace distribution.

        https://en.wikipedia.org/wiki/Laplace_distribution

        - Univariate continuous distribution
        - Support: :math:`\mathbb{R}`.
        - Parameters: location :math:`\\mu \in \mathbb{R}` and scale :math:`b > 0`.

        Parameters
        ----------
        prior_mu
            location :math:`\\mu` of the prior Laplace distribution (default: None).
        prior_b
            scale :math:`b` of the prior Laplace distribution (default: None).
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_mu = prior_mu
        self.prior_b = prior_b
        _check_strict_positive(self.prior_b)
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_mu, self.prior_b

    def _distr_from_params(self, params: Tuple):
        mu, b = params
        return _Laplace(mu, b)

    def sample(self, model_output):
        mu, b = self._params_from_output(model_output)
        distr = _Laplace(mu, b)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output: torch.Tensor):
        output_size = model_output.shape[-1]
        mu = model_output[:, :, :output_size // 2]
        b = self.softplus(model_output[:, :, output_size // 2:])
        return mu, b


class LogNormalLikelihood(Likelihood):
    def __init__(self, prior_mu=None, prior_sigma=None, prior_strength=1.):
        """
        Log-normal distribution.

        https://en.wikipedia.org/wiki/Log-normal_distribution

        - Univariate continuous distribution.
        - Support: :math:`\mathbb{R}_{>0}`.
        - Parameters: :math:`\\mu \in \mathbb{R}` and :math:`\\sigma > 0`.

        Parameters
        ----------
        prior_mu
            parameter :math:`\\mu` of the prior log-normal distribution (default: None).
        prior_sigma
            parameter :math:`\\sigma` of the prior log-normal distribution (default: None)
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.prior_mu = prior_mu
        self.prior_sigma = prior_sigma
        _check_strict_positive(self.prior_sigma, 'sigma')
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return self.prior_mu, self.prior_sigma

    def _distr_from_params(self, params):
        mu, sigma = params
        return _LogNormal(mu, sigma)

    def sample(self, model_output: torch.Tensor) -> torch.Tensor:
        mu, sigma = self._params_from_output(model_output)
        distr = _LogNormal(mu, sigma)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output):
        output_size = model_output.shape[-1]
        mu = model_output[:, :, :output_size // 2]
        sigma = self.softplus(model_output[:, :, output_size // 2:])
        return mu, sigma


class WeibullLikelihood(Likelihood):
    def __init__(self, prior_strength=1.):
        """
        Weibull distribution.

        https://en.wikipedia.org/wiki/Weibull_distribution

        - Univariate continuous distribution
        - Support: :math:`\mathbb{R}_{>0}`.
        - Parameters: scale :math:`\\lambda > 0` and concentration :math:`k > 0`.

        It does not support priors.

        Parameters
        ----------
        prior_strength
            strength of the loss regularisation induced by the prior
        """
        self.softplus = nn.Softplus()
        super().__init__(prior_strength)

    @property
    def _prior_params(self):
        return None

    def _distr_from_params(self, params: Tuple):
        lmba, k = params
        return _Weibull(lmba, k)

    def sample(self, model_output):
        lmbda, k = self._params_from_output(model_output)
        distr = _Weibull(lmbda, k)
        return distr.sample()

    @property
    def num_parameters(self) -> int:
        return 2

    def _params_from_output(self, model_output: torch.Tensor):
        output_size = model_output.shape[-1]
        lmbda = self.softplus(model_output[:, :, :output_size // 2])
        k = self.softplus(model_output[:, :, output_size // 2:])
        return lmbda, k


if False:
    """ TODO
        To make it work, we'll have to change our models so they optionally accept an absolute
        number of parameters, instead of num_perameters per component.
    """

    class MultivariateNormal(Likelihood):
        def __init__(self, dim: int, prior_mu=None, prior_covmat=None, prior_strength=1.):
            self.dim = dim
            self.prior_mu = prior_mu
            self.prior_covmat = prior_covmat
            if self.prior_mu is not None:
                raise_if_not(len(self.prior_mu) == self.dim, 'The provided prior_mu must have a size matching the '
                                                             'provided dimension.')
            if self.prior_covmat is not None:
                raise_if_not(self.prior_covmat.shape == (self.dim, self.dim), 'The provided prior on the covariaance '
                                                                              'matrix must have size (dim, dim).')
                _check_strict_positive(self.prior_covmat.flatten(), 'covariance matrix')

            self.softplus = nn.Softplus()
            super().__init__(prior_strength)

        @property
        def _prior_params(self):
            return self.prior_mu, self.prior_covmat

        def _distr_from_params(self, params: Tuple):
            mu, covmat = params
            return _MultivariateNormal(mu, covmat)

        def sample(self, model_output: torch.Tensor):
            mu, covmat = self._params_from_output(model_output)
            distr = _MultivariateNormal(mu, covmat)
            return distr.sample()

        @property
        def num_parameters(self) -> int:
            return int(self.dim + (self.dim**2 - self.dim) / 2)

        def _params_from_output(self, model_output: torch.Tensor):
            device = model_output.device
            mu = model_output[:, :, :self.dim]
            covmat_coefs = self.softplus(model_output[:, :, self.dim:])

            print('model output: {}'.format(model_output.shape))

            # build covariance matrix
            covmat = torch.zeros((model_output.shape[0], model_output.shape[1], self.dim, self.dim)).to(device)
            tril_indices = torch.tril_indices(row=self.dim, col=self.dim, offset=1, device=device)
            covmat[tril_indices[0], tril_indices[1]] = covmat_coefs
            covmat[tril_indices[1], tril_indices[0]] = covmat_coefs
            covmat[range(self.dim), range(self.dim)] = 1.




"""
Utils for time series generation
--------------------------------
"""

import math

from typing import Union, Optional

import numpy as np
import pandas as pd
import holidays

from ..timeseries import TimeSeries
from ..logging import raise_if_not, get_logger, raise_log, raise_if

logger = get_logger(__name__)


def _generate_index(start: Union[pd.Timestamp, int],
                    length: int = 10,
                    freq: str = 'D') -> Union[pd.DatetimeIndex, pd.Int64Index]:
    """Returns an index with a given start point and length. Either a pandas DatetimeIndex with given frequency
    or a pandas Int64Index. The index starts at

    Parameters
    ----------
    start
        The start of the returned index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    length
        The length of the returned index
    freq
        The time difference between two adjacent entries in the returned index. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    """

    if isinstance(start, pd.Timestamp):
        index = pd.date_range(start=start, periods=length, freq=freq)
    else:  # int
        index = pd.Int64Index(range(start, start + length))
    return index


def constant_timeseries(value: float = 1,
                        length: int = 10,
                        freq: str = 'D',
                        start: Union[pd.Timestamp, int] = pd.Timestamp('2000-01-01'),
                        column_name: Optional[str] = 'constant') -> TimeSeries:
    """
    Creates a constant univariate TimeSeries with the given value, length, start date and frequency.

    Parameters
    ----------
    value
        The constant value that the TimeSeries object will assume at every index.
    length
        The length of the returned TimeSeries.
    freq
        The time difference between two adjacent entries in the returned TimeSeries. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A constant TimeSeries with value 'value'.
    """

    index = _generate_index(start=start, freq=freq, length=length)
    values = np.full(length, value)

    return TimeSeries.from_times_and_values(index, values, freq=freq, columns=pd.Index([column_name]))


def linear_timeseries(start_value: float = 0,
                      end_value: float = 1,
                      length: int = 10,
                      freq: str = 'D',
                      start: Union[pd.Timestamp, int] = pd.Timestamp('2000-01-01'),
                      column_name: Optional[str] = 'linear') -> TimeSeries:
    """
    Creates a univariate TimeSeries with a starting value of `start_value` that increases linearly such that
    it takes on the value `end_value` at the last entry of the TimeSeries. This means that
    the difference between two adjacent entries will be equal to
    (`end_value` - `start_value`) / (`length` - 1).

    Parameters
    ----------
    start_value
        The value of the first entry in the TimeSeries.
    end_value
        The value of the last entry in the TimeSeries.
    length
        The length of the returned TimeSeries.
    freq
        The time difference between two adjacent entries in the returned TimeSeries. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A linear TimeSeries created as indicated above.
    """

    index = _generate_index(start=start, freq=freq, length=length)
    values = np.linspace(start_value, end_value, length)
    return TimeSeries.from_times_and_values(index, values, freq=freq, columns=pd.Index([column_name]))


def sine_timeseries(value_frequency: float = 0.1,
                    value_amplitude: float = 1.,
                    value_phase: float = 0.,
                    value_y_offset: float = 0.,
                    length: int = 10,
                    freq: str = 'D',
                    start: Union[pd.Timestamp, int] = pd.Timestamp('2000-01-01'),
                    column_name: Optional[str] = 'sine') -> TimeSeries:
    """
    Creates a univariate TimeSeries with a sinusoidal value progression with a given frequency, amplitude,
    phase and y offset.

    Parameters
    ----------
    value_frequency
        The number of periods that take place within one time unit given in `freq`.
    value_amplitude
        The maximum  difference between any value of the returned TimeSeries and `y_offset`.
    value_phase
        The relative position within one period of the first value of the returned TimeSeries (in radians).
    value_y_offset
        The shift of the sine function along the y axis.
    length
        The length of the returned TimeSeries.
    freq
        The time difference between two adjacent entries in the returned TimeSeries. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A sinusoidal TimeSeries parametrized as indicated above.
    """

    index = _generate_index(start=start, freq=freq, length=length)
    values = np.array(range(length), dtype=float)
    f = np.vectorize(
        lambda x: value_amplitude * math.sin(2 * math.pi * value_frequency * x + value_phase) + value_y_offset
    )
    values = f(values)

    return TimeSeries.from_times_and_values(index, values, freq=freq, columns=pd.Index([column_name]))


def gaussian_timeseries(length: int = 10,
                        freq: str = 'D',
                        mean: Union[float, np.ndarray] = 0.,
                        std: Union[float, np.ndarray] = 1.,
                        start: Union[pd.Timestamp, int] = pd.Timestamp('2000-01-01'),
                        column_name: Optional[str] = 'gaussian') -> TimeSeries:
    """
    Creates a gaussian univariate TimeSeries by sampling all the series values independently,
    from a gaussian distribution with mean `mean` and standard deviation `std`.

    Parameters
    ----------
    length
        The length of the returned TimeSeries.
    mean
        The mean of the gaussian distribution that is sampled at each step.
        If a float value is given, the same mean is used at every step.
        If a numpy.ndarray of floats with the same length as `length` is
        given, a different mean is used at each time step.
    std
        The standard deviation of the gaussian distribution that is sampled at each step.
        If a float value is given, the same standard deviation is used at every step.
        If an array of dimension `(length, length)` is given, it will
        be used as covariance matrix for a multivariate gaussian distribution.
    freq
        The time difference between two adjacent entries in the returned TimeSeries. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A white noise TimeSeries created as indicated above.
    """

    if type(mean) == np.ndarray:
        raise_if_not(mean.shape == (length,), 'If a vector of means is provided, '
                                              'it requires the same length as the TimeSeries.', logger)
    if type(std) == np.ndarray:
        raise_if_not(std.shape == (length, length), 'If a matrix of standard deviations is provided, '
                                                    'its shape has to match the length of the TimeSeries.', logger)

    index = _generate_index(start=start, freq=freq, length=length)
    values = np.random.normal(mean, std, size=length)

    return TimeSeries.from_times_and_values(index, values, freq=freq, columns=pd.Index([column_name]))


def random_walk_timeseries(length: int = 10,
                           freq: str = 'D',
                           mean: float = 0.,
                           std: float = 1.,
                           start: Union[pd.Timestamp, int] = pd.Timestamp('2000-01-01'),
                           column_name: Optional[str] = 'random_walk') -> TimeSeries:
    """
    Creates a random walk univariate TimeSeries, where each step is obtained by sampling a gaussian distribution
    with mean `mean` and standard deviation `std`.

    Parameters
    ----------
    length
        The length of the returned TimeSeries.
    mean
        The mean of the gaussian distribution that is sampled at each step.
    std
        The standard deviation of the gaussian distribution that is sampled at each step.
    freq
        The time difference between two adjacent entries in the returned TimeSeries. Only effective if `start` is a
        pandas Timestamp. A DateOffset alias is expected; see
        `docs <https://pandas.pydata.org/pandas-docs/stable/user_guide/TimeSeries.html#dateoffset-objects>`_.
    start
        The start of the returned TimeSeries' index. If a pandas Timestamp is passed, the TimeSeries will have a pandas
        DatetimeIndex. If an integer is passed, the TimeSeries will have a pandas Int64Index index.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A random walk TimeSeries created as indicated above.
    """

    index = _generate_index(start=start, freq=freq, length=length)
    values = np.cumsum(np.random.normal(mean, std, size=length))

    return TimeSeries.from_times_and_values(index, values, freq=freq, columns=pd.Index([column_name]))


def _extend_time_index_until(time_index: Union[pd.DatetimeIndex, pd.RangeIndex],
                             until: Optional[Union[int, str, pd.Timestamp]],
                             add_length: int,
                             ) -> pd.DatetimeIndex:

    if not add_length and not until:
        return time_index

    raise_if(bool(add_length) and bool(until), "set only one of add_length and until")

    end = time_index[-1]
    freq = time_index.freq

    if add_length:
        raise_if_not(add_length >= 0, f"Expected add_length, by which to extend the time series by, "
                                      f"to be positive, got {add_length}")

        try:
            end += add_length*freq
        except pd.errors.OutOfBoundsDatetime:
            raise_log(ValueError(f"the add operation between {end} and {add_length * freq} will overflow"), logger)
    else:
        datetime_index = isinstance(time_index, pd.DatetimeIndex)

        if datetime_index:
            raise_if_not(isinstance(until, (str, pd.Timestamp)), "Expected valid timestamp for TimeSeries, "
                                                                 "indexed by DatetimeIndex, "
                                                                 f"for parameter until, got {type(end)}", logger)
        else:
            raise_if_not(isinstance(until, int),  "Expected integer for TimeSeries,"
                                                  "indexed by RangeIndex, ",
                                                  f"for parameter until, got {type(end)}", logger)

        timestamp = pd.Timestamp(until) if datetime_index else until

        raise_if_not(timestamp > end, f"Expected until, {timestamp} to lie past end of time index {end}")

        ahead = timestamp - end
        raise_if_not((ahead % freq) == pd.Timedelta(0), f"End date must correspond with frequency {freq} of the time axis", logger)

        end = timestamp

    new_time_index = pd.date_range(start=time_index[0], end=end, freq=freq)
    return new_time_index


def holidays_timeseries(time_index: pd.DatetimeIndex,
                        country_code: str,
                        prov: str = None,
                        state: str = None,
                        column_name: Optional[str] = 'holidays',
                        until: Optional[Union[int, str, pd.Timestamp]] = None,
                        add_length: int = 0,
                        ) -> TimeSeries:
    """
    Creates a binary univariate TimeSeries with index `time_index` that equals 1 at every index that lies within
    (or equals) a selected country's holiday, and 0 otherwise.

    Available countries can be found `here <https://github.com/dr-prodigy/python-holidays#available-countries>`_.

    Parameters
    ----------
    time_index
        The time index over which to generate the holidays
    country_code
        The country ISO code
    prov
        The province
    state
        The state
    until
        Extend the time_index up until timestamp for datetime indexed series
        and int for range indexed series, should match or exceed forecasting window.
    add_length
        Extend the time_index by add_length, should match or exceed forecasting window.
        Set only one of until and add_length.
    column_name
        Optionally, the name of the value column for the returned TimeSeries

    Returns
    -------
    TimeSeries
        A new binary holiday TimeSeries instance.
    """

    time_index = _extend_time_index_until(time_index, until, add_length)
    scope = range(time_index[0].year, (time_index[-1] + pd.Timedelta(days=1)).year)
    country_holidays = holidays.CountryHoliday(country_code, prov=prov, state=state, years=scope)
    index_series = pd.Series(time_index, index=time_index)
    values = index_series.apply(lambda x: x in country_holidays).astype(int)
    return TimeSeries.from_times_and_values(time_index, values, columns=pd.Index([column_name]))


def datetime_attribute_timeseries(time_index: Union[pd.DatetimeIndex, TimeSeries],
                                  attribute: str,
                                  one_hot: bool = False,
                                  cyclic: bool = False,
                                  until: Optional[Union[int, str, pd.Timestamp]] = None,
                                  add_length: int = 0) -> TimeSeries:
    """
    Returns a new TimeSeries with index `time_index` and one or more dimensions containing
    (optionally one-hot encoded or cyclic encoded) pd.DatatimeIndex attribute information derived from the index.


    Parameters
    ----------
    time_index
        Either a `pd.DatetimeIndex` attribute which will serve as the basis of the new column(s), or
        a `TimeSeries` whose time axis will serve this purpose.
    attribute
        An attribute of `pd.DatetimeIndex`, or `week` / `weekofyear` / `week_of_year` - e.g. "month", "weekday", "day", "hour", "minute", "second". 
        See all available attributes in https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex .
    one_hot
        Boolean value indicating whether to add the specified attribute as a one hot encoding
        (results in more columns).
    cyclic
        Boolean value indicating whether to add the specified attribute as a cyclic encoding.
        Alternative to one_hot encoding, enable only one of the two.
        (adds 2 columns, corresponding to sin and cos transformation)
    until
        Extend the time_index up until timestamp for datetime indexed series
        and int for range indexed series, should match or exceed forecasting window.
    add_length
        Extend the time_index by add_length, should match or exceed forecasting window.
        Set only one of until and add_length.

    Returns
    -------
    TimeSeries
        New datetime attribute TimeSeries instance.
    """

    if isinstance(time_index, TimeSeries):
        time_index = time_index.time_index

    time_index = _extend_time_index_until(time_index, until, add_length)

    raise_if_not(
        hasattr(pd.DatetimeIndex, attribute) or (attribute in ["week", "weekofyear", "week_of_year"]),
        '"attribute" needs to be an attribute of pd.DatetimeIndex.'
        'See all available attributes in '
        'https://pandas.pydata.org/pandas-docs/stable/reference/api/pandas.DatetimeIndex.html#pandas.DatetimeIndex',
        logger,
    )

    raise_if(one_hot and cyclic, "set only one of one_hot or cyclic to true", logger)

    num_values_dict = {
        "month": 12,
        "day": 31,
        "weekday": 7,
        "dayofweek":7,
        "day_of_week":7,
        "hour": 24,
        "minute": 60,
        "second": 60,
        "microsecond": 1000000,
        "nanosecond": 1000,
        "quarter": 4,
        "dayofyear":365,
        "day_of_year":365,
        "week":52,
        "weekofyear":52,
        "week_of_year":52,
    }

    if not attribute in ["week", "weekofyear", "week_of_year"]:
        values = getattr(time_index, attribute)
    else:
        values = time_index.isocalendar().set_index("week").index.astype("int64").rename('time')
    
    if one_hot or cyclic:
        raise_if_not(
            attribute in num_values_dict,
            "Given datetime attribute not supported"
            " with one-hot or cyclical encoding."
            f"Supported datetime attribute: {list(num_values_dict.keys())}",
            logger,
        )

    if one_hot:
        values_df = pd.get_dummies(values)
        # fill missing columns (in case not all values appear in time_index)
        for i in range(1, num_values_dict[attribute] + 1):
            if not (i in values_df.columns):
                values_df[i] = 0
        values_df = values_df[range(1, num_values_dict[attribute] + 1)]

        values_df.columns = [
            attribute + "_" + str(column_name) for column_name in values_df.columns
        ]

    else:
        if cyclic:
            if attribute == "day":
                periods = [time_index[i].days_in_month for i in time_index.month]
                freq = 2 * np.pi * np.reciprocal(periods)
            else:
                period = num_values_dict[attribute]
                freq = 2 * np.pi / period

            values_df = pd.DataFrame(
                {
                    attribute + "_sin": np.sin(freq * values),
                    attribute + "_cos": np.cos(freq * values),
                }
            )
        else:
            values_df = pd.DataFrame({attribute: values})

    values_df.index = time_index

    return TimeSeries.from_dataframe(values_df)



import logging
import numpy as np
import pandas as pd
import torch
from sklearn.ensemble import RandomForestRegressor
from sklearn.linear_model import LinearRegression

from .base_test_class import DartsBaseTestClass
from ..utils import timeseries_generation as tg
from ..models import NaiveDrift, NaiveSeasonal
from ..models import RegressionEnsembleModel, LinearRegressionModel, RandomForest, RegressionModel
from ..logging import get_logger
from .test_ensemble_models import _make_ts
from ..metrics import rmse
from .. import TimeSeries
from .test_regression_models import train_test_split

logger = get_logger(__name__)

try:
    from ..models import RNNModel, BlockRNNModel
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not available. Some tests will be skipped.")
    TORCH_AVAILABLE = False


class RegressionEnsembleModelsTestCase(DartsBaseTestClass):

    RANDOM_SEED = 111

    sine_series = tg.sine_timeseries(value_frequency=(1 / 5), value_y_offset=10, length=50)
    lin_series = tg.linear_timeseries(length=50)

    combined = sine_series + lin_series

    seq1 = [_make_ts(0), _make_ts(10), _make_ts(20)]
    cov1 = [_make_ts(5), _make_ts(15), _make_ts(25)]

    seq2 = [_make_ts(0, 20), _make_ts(10, 20), _make_ts(20, 20)]
    cov2 = [_make_ts(5, 30), _make_ts(15, 30), _make_ts(25, 30)]

    # dummy feature and target TimeSeries instances
    ts_periodic = tg.sine_timeseries(length=500)
    ts_gaussian = tg.gaussian_timeseries(length=500)
    ts_random_walk = tg.random_walk_timeseries(length=500)

    ts_cov1 = ts_periodic.stack(ts_gaussian)
    ts_cov1 = ts_cov1.pd_dataframe()
    ts_cov1.columns = ["Periodic", "Gaussian"]
    ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
    ts_sum1 = ts_periodic + ts_gaussian

    ts_cov2 = ts_sum1.stack(ts_random_walk)
    ts_sum2 = ts_sum1 + ts_random_walk

    def get_local_models(self):
        return [NaiveDrift(), NaiveSeasonal(5), NaiveSeasonal(10)]

    def get_global_models(self, output_chunk_length=5):
        return [RNNModel(input_chunk_length=20, output_chunk_length=output_chunk_length, n_epochs=1,
                         random_state=42),
                BlockRNNModel(input_chunk_length=20, output_chunk_length=output_chunk_length, n_epochs=1,
                              random_state=42),
                ]

    def test_accepts_different_regression_models(self):
        regr1 = LinearRegression()
        regr2 = RandomForestRegressor()
        regr3 = RandomForest(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel(self.get_local_models(), 10)
        model1 = RegressionEnsembleModel(self.get_local_models(), 10, regr1)
        model2 = RegressionEnsembleModel(self.get_local_models(), 10, regr2)
        model3 = RegressionEnsembleModel(self.get_local_models(), 10, regr3)

        models = [model0, model1, model2, model3]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_accepts_one_model(self):
        regr1 = LinearRegression()
        regr2 = RandomForest(lags_future_covariates=[0])

        model0 = RegressionEnsembleModel([self.get_local_models()[0]], 10)
        model1 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr1)
        model2 = RegressionEnsembleModel([self.get_local_models()[0]], 10, regr2)

        models = [model0, model1, model2]
        for model in models:
            model.fit(series=self.combined)
            model.predict(10)

    def test_train_n_points(self):
        regr = LinearRegressionModel(lags_future_covariates=[0])

        # same values
        ensemble = RegressionEnsembleModel(self.get_local_models(), 5, regr)

        # too big value to perform the split
        ensemble = RegressionEnsembleModel(self.get_local_models(), 100)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

        ensemble = RegressionEnsembleModel(self.get_local_models(), 50)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

        # too big value considering min_train_series_length
        ensemble = RegressionEnsembleModel(self.get_local_models(), 45)
        with self.assertRaises(ValueError):
            ensemble.fit(self.combined)

    if TORCH_AVAILABLE:

        def test_torch_models_retrain(self):
            model1 = BlockRNNModel(input_chunk_length=12, output_chunk_length=1, random_state=0, n_epochs=2)
            model2 = BlockRNNModel(input_chunk_length=12, output_chunk_length=1, random_state=0, n_epochs=2)

            ensemble = RegressionEnsembleModel([model1], 5)
            ensemble.fit(self.combined)

            model1_fitted = ensemble.models[0]
            forecast1 = model1_fitted.predict(10)

            model2.fit(self.combined)
            forecast2 = model2.predict(10)

            self.assertAlmostEqual(sum(forecast1.values() - forecast2.values())[0], 0., places=2)

        def test_train_predict_global_models_univar(self):
            ensemble_models = self.get_global_models(output_chunk_length=10)
            ensemble_models.append(RegressionModel(lags=1))
            ensemble = RegressionEnsembleModel(ensemble_models, 10)
            ensemble.fit(series=self.combined)
            ensemble.predict(10)

        def test_train_predict_global_models_multivar_no_covariates(self):
            ensemble_models = self.get_global_models(output_chunk_length=10)
            ensemble_models.append(RegressionModel(lags=1))
            ensemble = RegressionEnsembleModel(ensemble_models, 10)
            ensemble.fit(self.seq1)
            ensemble.predict(10, self.seq1)

        def test_train_predict_global_models_multivar_with_covariates(self):
            ensemble_models = self.get_global_models(output_chunk_length=10)
            ensemble_models.append(RegressionModel(lags=1, lags_past_covariates=[-1]))
            ensemble = RegressionEnsembleModel(ensemble_models, 10)
            ensemble.fit(self.seq1, self.cov1)
            ensemble.predict(10, self.seq2, self.cov2)

        def helper_test_models_accuracy(self, model_instance, n, series, past_covariates, min_rmse):
            # for every model, test whether it predicts the target with a minimum r2 score of `min_rmse`
            train_f, train_t, test_f, test_t = train_test_split(past_covariates, series, pd.Timestamp("20010101"))

            model_instance.fit(series=train_t, past_covariates=train_f)
            prediction = model_instance.predict(n=n, past_covariates=past_covariates)
            current_rmse = rmse(prediction, test_t)

            self.assertTrue(
                current_rmse <= min_rmse,
                f"Model was not able to denoise data. A rmse score of {current_rmse} was recorded."
            )

        def denoising_input(self):
            np.random.seed(self.RANDOM_SEED)

            ts_periodic = tg.sine_timeseries(length=500)
            ts_gaussian = tg.gaussian_timeseries(length=500)
            ts_random_walk = tg.random_walk_timeseries(length=500)

            ts_cov1 = ts_periodic.stack(ts_gaussian)
            ts_cov1 = ts_cov1.pd_dataframe()
            ts_cov1.columns = ["Periodic", "Gaussian"]
            ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
            ts_sum1 = ts_periodic + ts_gaussian

            ts_cov2 = ts_sum1.stack(ts_random_walk)
            ts_sum2 = ts_sum1 + ts_random_walk

            return ts_sum1, ts_cov1, ts_sum2, ts_cov2

        def test_ensemble_models_denoising(self):
            # for every model, test whether it correctly denoises ts_sum using ts_gaussian and ts_sum as inputs
            horizon = 10
            ts_sum1, ts_cov1, _, _ = self.denoising_input()
            torch.manual_seed(self.RANDOM_SEED)

            ensemble_models = [
                RNNModel(input_chunk_length=20, output_chunk_length=horizon, n_epochs=1, random_state=self.RANDOM_SEED),
                BlockRNNModel(input_chunk_length=20, output_chunk_length=horizon, n_epochs=1, random_state=self.RANDOM_SEED),
                RegressionModel(lags=1, lags_past_covariates=[-1])
            ]

            ensemble = RegressionEnsembleModel(ensemble_models, horizon)
            self.helper_test_models_accuracy(ensemble, horizon, ts_sum1, ts_cov1, 1.5)

        def test_ensemble_models_denoising_multi_input(self):
            # for every model, test whether it correctly denoises ts_sum_2 using ts_random_multi and ts_sum_2 as inputs
            horizon = 10
            _, _, ts_sum2, ts_cov2 = self.denoising_input()
            torch.manual_seed(self.RANDOM_SEED)

            ensemble_models = [
                RNNModel(input_chunk_length=20, output_chunk_length=horizon, n_epochs=1, random_state=self.RANDOM_SEED),
                BlockRNNModel(input_chunk_length=20, output_chunk_length=horizon, n_epochs=1, random_state=self.RANDOM_SEED),
                RegressionModel(lags=1, lags_past_covariates=[-1]), RegressionModel(lags=1, lags_past_covariates=[-1])
            ]

            ensemble = RegressionEnsembleModel(ensemble_models, horizon)
            self.helper_test_models_accuracy(ensemble, horizon, ts_sum2, ts_cov2, 1.9)


import numpy as np
import pandas as pd

import darts
from darts import TimeSeries
from darts.metrics import rmse
from darts.models import RegressionModel, RandomForest, LinearRegressionModel, LightGBMModel
from .base_test_class import DartsBaseTestClass
from darts.utils import timeseries_generation as tg
from sklearn.linear_model import LinearRegression

# Required to import HistGradientBoostingRegressor in sklearn
from sklearn.experimental import enable_hist_gradient_boosting

from sklearn.ensemble import RandomForestRegressor, HistGradientBoostingRegressor
from darts.utils.data.sequential_dataset import MixedCovariatesSequentialDataset
from darts.utils.data.inference_dataset import MixedCovariatesInferenceDataset
from darts.models.forecasting.regression_model import _shift_matrices, _update_min_max
from unittest.mock import patch


def train_test_split(features, target, split_ts):
    """
    Splits all provided TimeSeries instances into train and test sets according to the provided timestamp.

    Parameters
    ----------
    features : TimeSeries
        Feature TimeSeries instances to be split.
    target : TimeSeries
        Target TimeSeries instance to be split.
    split_ts : TimeStamp
        Time stamp indicating split point.

    Returns
    -------
    TYPE
        4-tuple of the form (train_features, train_target, test_features, test_target)
    """
    train_features, test_features = features.split_after(split_ts)
    train_target, test_target = target.split_after(split_ts)

    return (train_features, train_target, test_features, test_target)


class RegressionModelsTestCase(DartsBaseTestClass):

    np.random.seed(42)

    # dummy feature and target TimeSeries instances
    ts_periodic = tg.sine_timeseries(length=500)
    ts_gaussian = tg.gaussian_timeseries(length=500)
    ts_random_walk = tg.random_walk_timeseries(length=500)

    ts_cov1 = ts_periodic.stack(ts_gaussian)
    ts_cov1 = ts_cov1.pd_dataframe()
    ts_cov1.columns = ["Periodic", "Gaussian"]
    ts_cov1 = TimeSeries.from_dataframe(ts_cov1)
    ts_sum1 = ts_periodic + ts_gaussian

    ts_cov2 = ts_sum1.stack(ts_random_walk)
    ts_sum2 = ts_sum1 + ts_random_walk

    # default regression models
    models = [
        RandomForest,
        LinearRegressionModel,
        RegressionModel,
        LightGBMModel
    ]

    target_series = tg.linear_timeseries(start_value=0, end_value=49, length=50)
    past_covariates = tg.linear_timeseries(start_value=50, end_value=99, length=50).stack(
        tg.linear_timeseries(start_value=50.5, end_value=99.5, length=50)
    )
    future_covariates = tg.linear_timeseries(start_value=100, end_value=199, length=100)

    lags_1 = [-5, -3, -1]
    lags_past_covariates_1 = [-4, -2]
    lags_future_covariates_1 = [-1, 0, 3]

    def test_model_construction(self):

        for model in self.models:
            # TESTING SINGLE INT
            # testing lags
            model_instance = model(lags=5)
            self.assertEqual(model_instance.lags, [-5, -4, -3, -2, -1])
            self.assertEqual(model_instance.min_lag, -5)
            self.assertEqual(model_instance.max_lag, -1)
            # testing lags_past_covariates
            model_instance = model(lags=None, lags_past_covariates=3)
            self.assertEqual(model_instance.lags_past_covariates, [-3, -2, -1])
            self.assertEqual(model_instance.min_lag, -3)
            self.assertEqual(model_instance.max_lag, -1)
            # testing lags_future covariates
            model_instance = model(lags=None, lags_future_covariates=(3, 5))
            self.assertEqual(model_instance.lags_historical_covariates, [-3, -2, -1])
            self.assertEqual(model_instance.lags_future_covariates, [0, 1, 2, 3, 4])
            self.assertEqual(model_instance.min_lag, -3)
            self.assertEqual(model_instance.max_lag, 4)

            # TESTING LIST of int
            # lags
            values = [-5, -3, -1]
            model_instance = model(lags=values)
            self.assertEqual(model_instance.lags, values)
            self.assertEqual(model_instance.min_lag, -5)
            self.assertEqual(model_instance.max_lag, -1)
            # testing lags_past_covariates
            model_instance = model(lags_past_covariates=values)
            self.assertEqual(model_instance.lags_past_covariates, values)
            self.assertEqual(model_instance.min_lag, -5)
            self.assertEqual(model_instance.max_lag, -1)
            # testing lags_future_covariates

            checks_future_covariates = [
                # lags as input, lags_historical_covariates, lags_future_covariates, min_lag, max_lag
                ([-3, -5, 1, 5], [-5, -3], [1, 5], -5, 5),
                ([-3, -5], [-5, -3], None, -5, -3),
                ([1, 5], None, [1, 5], 1, 5)
            ]

            for lags, lags_hist, lags_fut, min_l, max_l in checks_future_covariates:
                model_instance = model(lags_future_covariates=lags)
                self.assertEqual(model_instance.lags_historical_covariates, lags_hist)
                self.assertEqual(model_instance.lags_future_covariates, lags_fut)
                self.assertEqual(model_instance.min_lag, min_l)
                self.assertEqual(model_instance.max_lag, max_l)

            with self.assertRaises(ValueError):
                model()
            with self.assertRaises(ValueError):
                model(lags=0)
            with self.assertRaises(ValueError):
                model(lags=[-1, 0])
            with self.assertRaises(ValueError):
                model(lags=[3, 5])
            with self.assertRaises(ValueError):
                model(lags=[-3, -5.0])
            with self.assertRaises(ValueError):
                model(lags=-5)
            with self.assertRaises(ValueError):
                model(lags=3.6)
            with self.assertRaises(ValueError):
                model(lags=None, lags_past_covariates=False)
            with self.assertRaises(ValueError):
                model(lags=None)
            with self.assertRaises(ValueError):
                model(lags=5, lags_future_covariates=True)
            with self.assertRaises(ValueError):
                model(lags=5, lags_future_covariates=(1, -3))
            with self.assertRaises(ValueError):
                model(lags=5, lags_future_covariates=(1, 2, 3))
            with self.assertRaises(ValueError):
                model(lags=5, lags_future_covariates=(1, True))
            with self.assertRaises(ValueError):
                model(lags=5, lags_future_covariates=(1, 1.0))

    def test_training_data_creation(self):
        # testing _get_training_data function
        model_instance = RegressionModel(
            lags=self.lags_1,
            lags_past_covariates=self.lags_past_covariates_1,
            lags_future_covariates=self.lags_future_covariates_1
        )

        input_chunk_length = -model_instance.min_lag
        training_output_chunk_length = max(model_instance.max_lag + 1, 1)

        training_dataset = MixedCovariatesSequentialDataset(
            target_series=self.target_series,
            past_covariates=self.past_covariates,
            future_covariates=self.future_covariates,
            input_chunk_length=input_chunk_length,
            output_chunk_length=training_output_chunk_length
        )

        training_samples, training_labels = model_instance._get_training_data(training_dataset=training_dataset)

        # checking number of dimensions
        self.assertEqual(len(training_samples.shape), 2)
        self.assertEqual(len(training_labels.shape), 1)
        self.assertEqual(training_samples.shape[0], training_labels.shape[0])
        self.assertEqual(training_samples.shape[1],
                         len(self.lags_1) + len(self.lags_past_covariates_1) * 2 + len(self.lags_future_covariates_1))

        # checking first row order lags | lags_past_cov | lags_future_cov
        self.assertListEqual(list(training_samples[-1, :]), [0, 2, 4, 51, 51.5, 53, 53.5, 104, 105, 108])
        self.assertEqual(training_labels[-1], 5)

    def test_prediction_data_creation(self):
        model_instance = RegressionModel(
            lags=self.lags_1,
            lags_past_covariates=self.lags_past_covariates_1,
            lags_future_covariates=self.lags_future_covariates_1
        )

        n = 2
        input_chunk_length = max(1, -model_instance.min_lag)
        prediction_output_chunk_length = max(n + model_instance.max_lag, n)

        prediction_dataset = MixedCovariatesInferenceDataset(
            target_series=self.target_series,
            past_covariates=self.past_covariates,
            future_covariates=self.future_covariates,
            input_chunk_length=input_chunk_length,
            output_chunk_length=prediction_output_chunk_length
        )

        (
            target_matrix,
            past_covariates_matrix,
            historic_future_covariates_matrix,
            future_covariates_matrix,
            future_past_covariates_matrix
        ) = model_instance._get_prediction_data(prediction_dataset)

        # checking matrices sizes and content

        for matrix in [target_matrix, past_covariates_matrix, historic_future_covariates_matrix,
                       future_covariates_matrix]:
            self.assertEqual(matrix.shape[0], 1)  # we are training on a single ts

        # checking matrices dimensions
        self.assertEqual(target_matrix.ndim, 2)  # samples, time (univariate)
        self.assertEqual(past_covariates_matrix.ndim, 3)  # samples, time, dim (could be multivariate)
        self.assertEqual(historic_future_covariates_matrix.ndim, 3)  # same
        self.assertEqual(future_covariates_matrix.ndim, 3)  # same

        # checking the first sample of the matrices, (should contain input_chunk_length matrices) for
        # past values, while should contain n + max_lag in the future ones (where not empty)
        np.testing.assert_array_equal(target_matrix[0].ravel(), np.arange(45, 50))
        np.testing.assert_array_equal(past_covariates_matrix[0, :, 0], np.arange(95, 100))
        np.testing.assert_array_equal(past_covariates_matrix[0, :, 1], np.arange(95.5, 100.5))
        np.testing.assert_array_equal(historic_future_covariates_matrix[0].ravel(), np.arange(145, 150))
        self.assertIsNone(future_past_covariates_matrix)
        np.testing.assert_array_equal(future_covariates_matrix[0].ravel(), np.arange(150, 151))

    def test_shift_matrices(self):
        a = np.zeros((10, 8, 2))
        b = np.ones((10, 3, 2))
        a, b = _shift_matrices(a, b)
        # last "temporal slice" of a should now contain 1s
        np.testing.assert_array_equal(a[:, -1, :], np.ones((10, 2)))

        # testing empty future_matrix
        with self.assertRaises(ValueError):
            a, b = _shift_matrices(a, None)

    def test_update_min_max(self):
        current_min = None
        current_max = None
        values = [-1, -1, -4, 5, 9, 12]
        current_min, current_max = _update_min_max(current_min, current_max, values)

        self.assertEqual(current_min, -4)
        self.assertEqual(current_max, 12)

    def test_models_runnability(self):
        train_x, test_x = self.ts_cov1.split_before(0.7)
        train_y, test_y = self.ts_sum1.split_before(0.7)
        for model in self.models:
            # testing past covariates
            with self.assertRaises(ValueError):
                # testing lags_past_covariates None but past_covariates during training
                model_instance = model(lags=4, lags_past_covariates=None)
                model_instance.fit(series=self.ts_sum1, past_covariates=self.ts_cov1)

            with self.assertRaises(ValueError):
                # testing lags_past_covariates but no past_covariates during fit
                model_instance = model(lags=4, lags_past_covariates=3)
                model_instance.fit(series=self.ts_sum1)

            # testing future_covariates
            with self.assertRaises(ValueError):
                # testing lags_future_covariates None but future_covariates during training
                model_instance = model(lags=4, lags_future_covariates=None)
                model_instance.fit(series=self.ts_sum1, future_covariates=self.ts_cov1)

            with self.assertRaises(ValueError):
                # testing lags_covariate but no covariate during fit
                model_instance = model(lags=4, lags_future_covariates=3)
                model_instance.fit(series=self.ts_sum1)

            # testing input_dim
            model_instance = model(lags=4, lags_past_covariates=2)
            model_instance.fit(
                series=train_y,
                past_covariates=self.ts_sum1.stack(self.ts_sum1)
            )

            self.assertEqual(model_instance.input_dim, 3)

            with self.assertRaises(ValueError):
                prediction = model_instance.predict(n=len(test_y) + 2)

            # while it should work with n = 1
            prediction = model_instance.predict(n=1)
            self.assertTrue(
                len(prediction) == 1,
                f"Expected length 1, found {len(prediction)} instead",
            )

    def test_fit(self):
        for model in self.models:
            with self.assertRaises(ValueError):
                model_instance = model(lags=4, lags_past_covariates=4)
                model_instance.fit(series=self.ts_sum1, past_covariates=self.ts_cov1)
                model_instance.predict(n=10)

            model_instance = model(lags=12)
            model_instance.fit(series=self.ts_sum1)
            self.assertEqual(model_instance.lags_past_covariates, None)

            model_instance = model(lags=12, lags_past_covariates=12)
            model_instance.fit(series=self.ts_sum1, past_covariates=self.ts_cov1)
            self.assertEqual(len(model_instance.lags_past_covariates), 12)

            model_instance = model(lags=12, lags_future_covariates=(0, 1))
            model_instance.fit(series=self.ts_sum1, future_covariates=self.ts_cov1)
            self.assertEqual(len(model_instance.lags_future_covariates), 1)
            self.assertIsNone(model_instance.lags_historical_covariates)

            model_instance = model(lags=12, lags_past_covariates=[-1, -4, -6])
            model_instance.fit(series=self.ts_sum1, past_covariates=self.ts_cov1)
            self.assertEqual(len(model_instance.lags_past_covariates), 3)

            model_instance = model(lags=12, lags_past_covariates=[-1, -4, -6], lags_future_covariates=[-2, 0])
            model_instance.fit(series=self.ts_sum1, past_covariates=self.ts_cov1, future_covariates=self.ts_cov1)
            self.assertEqual(len(model_instance.lags_past_covariates), 3)

    def helper_test_models_accuracy(self, series, past_covariates, min_rmse):
        # for every model, test whether it predicts the target with a minimum r2 score of `min_rmse`
        train_f, train_t, test_f, test_t = train_test_split(past_covariates, series, pd.Timestamp("20010101"))

        for model in self.models:
            model_instance = model(lags=12, lags_past_covariates=2)
            model_instance.fit(series=train_t, past_covariates=train_f)
            prediction = model_instance.predict(n=len(test_t), past_covariates=past_covariates)
            current_rmse = rmse(prediction, test_t)

            self.assertTrue(
                current_rmse <= min_rmse,
                f"{str(model_instance)} model was not able to denoise data. A rmse score of {current_rmse} was"
                "recorded."
            )

    def test_models_denoising(self):
        # for every model, test whether it correctly denoises ts_sum using ts_gaussian and ts_sum as inputs
        self.helper_test_models_accuracy(self.ts_sum1, self.ts_cov1, 1.5)

    def test_models_denoising_multi_input(self):
        # for every model, test whether it correctly denoises ts_sum_2 using ts_random_multi and ts_sum_2 as inputs
        self.helper_test_models_accuracy(self.ts_sum2, self.ts_cov2, 19.5)

    def test_historical_forecast(self):
        model = self.models[0](lags=5)
        result = model.historical_forecasts(
            series=self.ts_sum1[:100],
            future_covariates=None,
            start=0.5,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False
        )
        self.assertEqual(len(result), 51)

        model = self.models[0](lags=5, lags_past_covariates=5)
        result = model.historical_forecasts(
            series=self.ts_sum1[:100],
            past_covariates=self.ts_cov1[:100],
            start=0.5,
            forecast_horizon=1,
            stride=1,
            retrain=True,
            overlap_end=False,
            last_points_only=True,
            verbose=False
        )
        self.assertEqual(len(result), 51)

    def test_regression_model(self):
        lags = 12
        models = [
            RegressionModel(lags=lags),
            RegressionModel(lags=lags, model=LinearRegression()),
            RegressionModel(lags=lags, model=RandomForestRegressor()),
            RegressionModel(lags=lags, model=HistGradientBoostingRegressor())
        ]

        for model in models:
            model.fit(series=self.ts_sum1)
            self.assertEqual(len(model.lags), lags)
            model.predict(n=10)

    def test_multiple_ts(self):
        lags = 4
        lags_past_covariates = 3
        model = RegressionModel(lags=lags, lags_past_covariates=lags_past_covariates)

        target_series = tg.linear_timeseries(start_value=0, end_value=49, length=50)
        past_covariates = tg.linear_timeseries(start_value=100, end_value=149, length=50)
        past_covariates = past_covariates.stack(
            tg.linear_timeseries(start_value=400, end_value=449, length=50))

        target_train, target_test = target_series.split_after(0.7)
        past_covariates_train, past_covariates_test = past_covariates.split_after(0.7)
        model.fit(
            series=[target_train, target_train + 0.5],
            past_covariates=[past_covariates_train, past_covariates_train + 0.5])

        predictions = model.predict(
            10,
            series=[target_train, target_train + 0.5],
            past_covariates=[past_covariates, past_covariates + 0.5])

        self.assertEqual(len(predictions[0]), 10, f"Found {len(predictions)} instead")

        # multiple TS, both future and past covariates, checking that both covariates lead to better results than using
        # a single one (target series = past_cov + future_cov + noise)
        np.random.seed(42)

        linear_ts_1 = tg.linear_timeseries(start_value=10, end_value=59, length=50)
        linear_ts_2 = tg.linear_timeseries(start_value=40, end_value=89, length=50)

        past_covariates = tg.sine_timeseries(length=50) * 10
        future_covariates = tg.sine_timeseries(length=50, value_frequency=0.015) * 50

        target_series_1 = linear_ts_1 + 4 * past_covariates + 2 * future_covariates
        target_series_2 = linear_ts_2 + 4 * past_covariates + 2 * future_covariates

        target_series_1_noise = (linear_ts_1 + 4 * past_covariates + 2 * future_covariates
                                 + tg.gaussian_timeseries(std=7, length=50))

        target_series_2_noise = (linear_ts_2 + 4 * past_covariates + 2 * future_covariates
                                 + tg.gaussian_timeseries(std=7, length=50))

        target_train_1, target_test_1 = target_series_1.split_after(0.7)
        target_train_2, target_test_2 = target_series_2.split_after(0.7)

        target_train_1_noise, target_test_1_noise = target_series_1_noise.split_after(0.7)
        target_train_2_noise, target_test_2_noise = target_series_2_noise.split_after(0.7)

        # testing improved denoise with multiple TS

        # test 1: with single TS, 2 covariates should be better than one
        model = RegressionModel(lags=3, lags_past_covariates=5)
        model.fit([target_train_1_noise], [past_covariates])

        prediction_past_only = model.predict(
            n=len(target_test_1), series=[target_train_1_noise, target_train_2_noise],
            past_covariates=[past_covariates] * 2
        )

        model = RegressionModel(lags=3, lags_past_covariates=5, lags_future_covariates=(5, 0))
        model.fit([target_train_1_noise], [past_covariates], [future_covariates])
        prediction_past_and_future = model.predict(
            n=len(target_test_1), series=[target_train_1_noise, target_train_2_noise],
            past_covariates=[past_covariates] * 2, future_covariates=[future_covariates] * 2
        )

        error_past_only = rmse([target_test_1, target_test_2], prediction_past_only, inter_reduction=np.mean)
        error_both = rmse([target_test_1, target_test_2], prediction_past_and_future, inter_reduction=np.mean)

        self.assertTrue(error_past_only > error_both)
        # test 2: with both covariates, 2 TS should learn more than one (with little noise)
        model = RegressionModel(lags=3, lags_past_covariates=5, lags_future_covariates=(5, 0))
        model.fit([target_train_1_noise, target_train_2_noise], [past_covariates] * 2, [future_covariates] * 2)
        prediction_past_and_future_multi_ts = model.predict(
            n=len(target_test_1), series=[target_train_1_noise, target_train_2_noise],
            past_covariates=[past_covariates] * 2, future_covariates=[future_covariates] * 2
        )
        error_both_multi_ts = rmse([target_test_1, target_test_2], prediction_past_and_future_multi_ts,
                                   inter_reduction=np.mean)

        self.assertTrue(error_both > error_both_multi_ts)

    def test_only_future_covariates(self):

        model = RegressionModel(lags_future_covariates=[-2])

        target_series = tg.linear_timeseries(start_value=0, end_value=49, length=50)
        covariates = tg.linear_timeseries(start_value=100, end_value=149, length=50)
        covariates = covariates.stack(
            tg.linear_timeseries(start_value=400, end_value=449, length=50)
        )

        target_train, target_test = target_series.split_after(0.7)
        covariates_train, covariates_test = covariates.split_after(0.7)
        model.fit(
            series=[target_train, target_train + 0.5],
            future_covariates=[covariates_train, covariates_train + 0.5],
        )

        predictions = model.predict(
            10,
            series=[target_train, target_train + 0.5],
            future_covariates=[covariates, covariates + 0.5],
        )

        self.assertEqual(len(predictions[0]), 10, f"Found {len(predictions[0])} instead")

    def test_not_enough_future_covariate(self):

        target_series = tg.linear_timeseries(start_value=0, end_value=19, length=20)
        covariates = tg.linear_timeseries(start_value=0, end_value=20, length=20)

        target_train, target_test = target_series.split_after(9)
        covariates_train, covariates_test = covariates.split_after(9)

        model = RegressionModel(lags_future_covariates=[0])
        model.fit(series=target_train, future_covariates=covariates_train)
        # 11 future covariates, with 0 lags covariate, can predict up to 10
        model.predict(10, series=target_train, future_covariates=covariates)
        with self.assertRaises(ValueError):
            model.predict(11, series=target_train, future_covariates=covariates)

        model = RegressionModel(lags_past_covariates=1)
        model.fit(series=target_train, past_covariates=covariates_train)
        # 11 future covariates, without 0 lags covariate, can predict up to 11
        model.predict(11, series=target_train, past_covariates=covariates)
        with self.assertRaises(ValueError):
            model.predict(12, series=target_train, past_covariates=covariates)

    @patch.object(darts.models.forecasting.gradient_boosted_model.lgb.LGBMRegressor, "fit")
    # @patch.object(darts.models.forecasting.gradient_boosted_model.lgb.LGBMRegressor, 'fit')
    def test_gradient_boosted_model_with_eval_set(self, lgb_fit_patch):
        """test whether these evaluation set parameters are passed to LGBRegressor """
        model = LightGBMModel(lags=4, lags_past_covariates=2)
        split_index = 450
        model.fit(series=self.ts_sum1[:split_index],
                  past_covariates=self.ts_cov1[:split_index],
                  val_series=self.ts_sum1[split_index:],
                  val_past_covariates=self.ts_cov1[split_index:],
                  early_stopping_rounds=2,
                  )

        lgb_fit_patch.assert_called_once()

        assert lgb_fit_patch.call_args[1]['eval_set'] is not None
        assert lgb_fit_patch.call_args[1]['early_stopping_rounds'] == 2


import unittest
import numpy as np
import pandas as pd
import random

from darts import TimeSeries
from darts.metrics import mape, r2_score
from darts.utils.timeseries_generation import (
    linear_timeseries as lt,
    sine_timeseries as st,
    random_walk_timeseries as rt,
    constant_timeseries as ct,
    gaussian_timeseries as gt
)
from darts.models import (
    Theta,
    FFT,
    ExponentialSmoothing,
    NaiveSeasonal,
    LinearRegressionModel,
    NaiveDrift,
    RandomForest,
    ARIMA
)


from .base_test_class import DartsBaseTestClass
from ..logging import get_logger

logger = get_logger(__name__)

try:
    from ..models import TCNModel, BlockRNNModel
    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch models are not installed - will not be tested for backtesting")
    TORCH_AVAILABLE = False


def compare_best_against_random(model_class, params, series, stride=1):

    # instantiate best model in expanding window mode
    np.random.seed(1)
    best_model_1, _ = model_class.gridsearch(params,
                                             series,
                                             forecast_horizon=10,
                                             stride=stride,
                                             metric=mape,
                                             start=series.time_index[-21])

    # instantiate best model in split mode
    train, val = series.split_before(series.time_index[-10])
    best_model_2, _ = model_class.gridsearch(params, train, val_series=val, metric=mape)

    # intantiate model with random parameters from 'params'
    random.seed(1)
    random_param_choice = {}
    for key in params.keys():
        random_param_choice[key] = random.choice(params[key])
    random_model = model_class(**random_param_choice)

    # perform backtest forecasting on both models
    best_score_1 = best_model_1.backtest(series, start=series.time_index[-21], forecast_horizon=10)
    random_score_1 = random_model.backtest(series, start=series.time_index[-21], forecast_horizon=10)

    # perform train/val evaluation on both models
    best_model_2.fit(train)
    best_score_2 = mape(best_model_2.predict(len(val)), series)
    random_model = model_class(**random_param_choice)
    random_model.fit(train)
    random_score_2 = mape(random_model.predict(len(val)), series)

    # check whether best models are at least as good as random models
    expanding_window_ok = best_score_1 <= random_score_1
    split_ok = best_score_2 <= random_score_2

    return expanding_window_ok and split_ok


class BacktestingTestCase(DartsBaseTestClass):
    def test_backtest_forecasting(self):
        linear_series = lt(length=50)
        linear_series_int = TimeSeries.from_values(linear_series.values())
        linear_series_multi = linear_series.stack(linear_series)

        # univariate model + univariate series
        score = NaiveDrift().backtest(linear_series, start=pd.Timestamp('20000201'),
                                      forecast_horizon=3, metric=r2_score)
        self.assertEqual(score, 1.0)

        # test that it also works for time series that are not Datetime-indexed
        score = NaiveDrift().backtest(linear_series_int, start=0.7,
                                      forecast_horizon=3, metric=r2_score)
        self.assertEqual(score, 1.0)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=pd.Timestamp('20000217'), forecast_horizon=3,
                                  overlap_end=False)
        NaiveDrift().backtest(linear_series, start=pd.Timestamp('20000216'), forecast_horizon=3)
        NaiveDrift().backtest(linear_series,
                              start=pd.Timestamp('20000217'),
                              forecast_horizon=3,
                              overlap_end=True)

        # Using forecast_horizon default value
        NaiveDrift().backtest(linear_series, start=pd.Timestamp("20000216"))
        NaiveDrift().backtest(linear_series, start=pd.Timestamp("20000217"), overlap_end=True)

        # Using an int or float value for start
        NaiveDrift().backtest(linear_series, start=30)
        NaiveDrift().backtest(linear_series, start=0.7, overlap_end=True)

        # Using invalid start and/or forecast_horizon values
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=0.7, forecast_horizon=-1)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=-0.7, forecast_horizon=1)

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=100)
        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=1.2)
        with self.assertRaises(TypeError):
            NaiveDrift().backtest(linear_series, start="wrong type")

        with self.assertRaises(ValueError):
            NaiveDrift().backtest(linear_series, start=49, forecast_horizon=2, overlap_end=False)

        # univariate model + multivariate series
        with self.assertRaises(AssertionError):
            NaiveDrift().backtest(linear_series_multi, start=pd.Timestamp("20000201"), forecast_horizon=3)

        # multivariate model + univariate series
        if TORCH_AVAILABLE:
            tcn_model = TCNModel(input_chunk_length=12, output_chunk_length=1, batch_size=1, n_epochs=1)
            pred = tcn_model.historical_forecasts(linear_series,
                                                  start=pd.Timestamp('20000125'),
                                                  forecast_horizon=3,
                                                  verbose=False,
                                                  last_points_only=True)
            self.assertEqual(pred.width, 1)
            self.assertEqual(pred.end_time(), linear_series.end_time())

            # multivariate model + multivariate series
            with self.assertRaises(ValueError):
                tcn_model.backtest(linear_series_multi, start=pd.Timestamp('20000125'),
                                   forecast_horizon=3, verbose=False)

            tcn_model = TCNModel(input_chunk_length=12, output_chunk_length=3, batch_size=1, n_epochs=1)
            pred = tcn_model.historical_forecasts(linear_series_multi,
                                                  start=pd.Timestamp('20000125'),
                                                  forecast_horizon=3,
                                                  verbose=False,
                                                  last_points_only=True)
            self.assertEqual(pred.width, 2)
            self.assertEqual(pred.end_time(), linear_series.end_time())

    def test_backtest_regression(self):
        gaussian_series = gt(mean=2, length=50)
        sine_series = st(length=50)
        features = gaussian_series.stack(sine_series)
        features_multivariate = (gaussian_series + sine_series).stack(gaussian_series).stack(sine_series)
        target = sine_series

        features = (
            features
            .with_columns_renamed(features.components, ["Value0", "Value1"])
        )

        features_multivariate = (
            features_multivariate
            .with_columns_renamed(features_multivariate.components, ["Value0", "Value1", "Value2"])
        )

        # univariate feature test
        score = LinearRegressionModel(lags=None, lags_future_covariates=[0, -1]).backtest(
            series=target, future_covariates=features, start=pd.Timestamp('20000201'),
            forecast_horizon=3, metric=r2_score, last_points_only=True
        )
        self.assertGreater(score, 0.9)

        # Using an int or float value for start
        score = RandomForest(lags=12, lags_future_covariates=[0], random_state=0).backtest(
            series=target, future_covariates=features, start=30,
            forecast_horizon=3, metric=r2_score
        )
        self.assertGreater(score, 0.9)

        score = RandomForest(lags=12, lags_future_covariates=[0], random_state=0).backtest(
            series=target, future_covariates=features, start=0.5,
            forecast_horizon=3, metric=r2_score
        )
        self.assertGreater(score, 0.9)

        # Using a too small start value
        with self.assertRaises(ValueError):
            RandomForest(lags=12).backtest(series=target, start=0, forecast_horizon=3)

        with self.assertRaises(ValueError):
            RandomForest(lags=12).backtest(
                series=target, start=0.01, forecast_horizon=3
            )

        # Using RandomForest's start default value
        score = RandomForest(lags=12, random_state=0).backtest(series=target, forecast_horizon=3, metric=r2_score)
        self.assertGreater(score, 0.95)

        # multivariate feature test
        score = RandomForest(lags=12, lags_future_covariates=[0, -1], random_state=0).backtest(
            series=target, future_covariates=features_multivariate,
            start=pd.Timestamp('20000201'), forecast_horizon=3, metric=r2_score
        )
        self.assertGreater(score, 0.94)

        # multivariate with stride
        score = RandomForest(lags=12, lags_future_covariates=[0], random_state=0).backtest(
            series=target, future_covariates=features_multivariate,
            start=pd.Timestamp('20000201'), forecast_horizon=3, metric=r2_score,
            last_points_only=True, stride=3
        )
        self.assertGreater(score, 0.9)

    def test_gridsearch(self):
        np.random.seed(1)

        ts_length = 50
        dummy_series = (
            lt(length=ts_length, end_value=10) + st(length=ts_length, value_y_offset=10) + rt(length=ts_length)
        )
        dummy_series_int_index = TimeSeries.from_values(dummy_series.values())

        theta_params = {"theta": list(range(3, 10))}
        self.assertTrue(compare_best_against_random(Theta, theta_params, dummy_series))
        self.assertTrue(compare_best_against_random(Theta, theta_params, dummy_series_int_index))
        self.assertTrue(compare_best_against_random(Theta, theta_params, dummy_series, stride=2))

        fft_params = {"nr_freqs_to_keep": [10, 50, 100], "trend": [None, "poly", "exp"]}
        self.assertTrue(compare_best_against_random(FFT, fft_params, dummy_series))

        es_params = {"seasonal_periods": list(range(5, 10))}
        self.assertTrue(
            compare_best_against_random(ExponentialSmoothing, es_params, dummy_series)
        )

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_n_jobs(self):
        """
        Testing that running gridsearch with multiple workers returns the same best_parameters as the single worker run.
        """

        np.random.seed(1)
        ts_length = 100

        dummy_series = (
            lt(length=ts_length, end_value=1) + st(length=ts_length, value_y_offset=0) + rt(length=ts_length)
        )

        ts_train = dummy_series[:round(ts_length * 0.8)]
        ts_val = dummy_series[round(ts_length * 0.8):]

        test_cases = [
            {
                "model": ARIMA,  # ExtendedForecastingModel
                "parameters": {"p": [18, 4, 8], "q": [1, 2, 3]}
            },
            {
                "model": BlockRNNModel,   # TorchForecastingModel
                "parameters": {
                    "input_chunk_length": [1, 3, 5, 10],
                    "output_chunk_length": [1, 3, 5, 10],
                    "n_epochs": [1, 5],
                    "random_state": [42]  # necessary to avoid randomness among runs with same parameters
                }
            }
        ]

        for test in test_cases:

            model = test["model"]
            parameters = test["parameters"]

            np.random.seed(1)
            _, best_params1 = model.gridsearch(parameters=parameters,
                                               series=ts_train,
                                               val_series=ts_val,
                                               n_jobs=1)

            np.random.seed(1)
            _, best_params2 = model.gridsearch(parameters=parameters,
                                               series=ts_train,
                                               val_series=ts_val,
                                               n_jobs=-1)

            self.assertEqual(best_params1, best_params2)

    @unittest.skipUnless(TORCH_AVAILABLE, "requires torch")
    def test_gridsearch_multi(self):
        dummy_series = st(length=40, value_y_offset=10).stack(lt(length=40, end_value=20))
        tcn_params = {
            "input_chunk_length": [12],
            "output_chunk_length": [3],
            "n_epochs": [1],
            "batch_size": [1],
            "kernel_size": [2, 3, 4],
        }
        TCNModel.gridsearch(tcn_params, dummy_series, forecast_horizon=3, metric=mape)

    def test_forecasting_residuals(self):
        model = NaiveSeasonal(K=1)

        # test zero residuals
        constant_ts = ct(length=20)
        residuals = model.residuals(constant_ts)
        np.testing.assert_almost_equal(
            residuals.univariate_values(), np.zeros(len(residuals))
        )

        # test constant, positive residuals
        linear_ts = lt(length=20)
        residuals = model.residuals(linear_ts)
        np.testing.assert_almost_equal(np.diff(residuals.univariate_values()), np.zeros(len(residuals) - 1))
        np.testing.assert_array_less(np.zeros(len(residuals)), residuals.univariate_values())



import pandas as pd
import numpy as np

from .base_test_class import DartsBaseTestClass
from ..timeseries import TimeSeries
from ..utils.data import (PastCovariatesInferenceDataset, FutureCovariatesInferenceDataset,
                          DualCovariatesInferenceDataset, MixedCovariatesInferenceDataset,
                          SplitCovariatesInferenceDataset, PastCovariatesSequentialDataset,
                          FutureCovariatesSequentialDataset, DualCovariatesSequentialDataset,
                          MixedCovariatesSequentialDataset, SplitCovariatesSequentialDataset,
                          PastCovariatesShiftedDataset, FutureCovariatesShiftedDataset,
                          DualCovariatesShiftedDataset, MixedCovariatesShiftedDataset,
                          SplitCovariatesShiftedDataset, HorizonBasedDataset)
from ..utils.timeseries_generation import gaussian_timeseries


class DatasetTestCase(DartsBaseTestClass):
    target1, target2 = gaussian_timeseries(length=100), gaussian_timeseries(length=150)
    vals1, vals2 = target1.values(), target2.values()
    cov1, cov2 = gaussian_timeseries(length=100), gaussian_timeseries(length=150)

    def _assert_eq(self, tup_ar, tup_series):
        l1 = []
        for ar_element in tup_ar:
            l1.append(None if ar_element is None else list(ar_element))
        l2 = []
        for series_element in tup_series:
            l2.append(None if series_element is None else list(series_element.values(copy=False)))

        self.assertEqual(l1, l2)

    def test_past_covariates_inference_dataset(self):
        # one target series
        ds = PastCovariatesInferenceDataset(target_series=self.target1, input_chunk_length=len(self.target1))
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self.assertEqual(ds[0][1:], (None, None, self.target1))

        # two target series
        ds = PastCovariatesInferenceDataset(target_series=[self.target1, self.target2],
                                            input_chunk_length=max(len(self.target1), len(self.target2)))
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self.assertEqual(ds[1][1:], (None, None, self.target2))

        # fail if covariates do not have same size
        with self.assertRaises(ValueError):
            ds = PastCovariatesInferenceDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # with covariates
        ds = PastCovariatesInferenceDataset(target_series=[self.target1, self.target2],
                                            covariates=[self.cov1, self.cov2],
                                            input_chunk_length=max(len(self.target1), len(self.target2)))
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        np.testing.assert_almost_equal(ds[1][1], self.cov2.values())
        self.assertEqual(ds[1][2:], (None, self.target2))  # no "future past" covariate here

        # more complex case with future past covariates:
        times1 = pd.date_range(start='20100101', end='20100701', freq='D')
        times2 = pd.date_range(start='20100101', end='20100820', freq='D')  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        short_cov = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        long_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = PastCovariatesInferenceDataset(target_series=target, covariates=short_cov,
                                            input_chunk_length=10, output_chunk_length=10, n=30)

        # should fail if covariates are too short
        with self.assertRaises(ValueError):
            ds[0]

        # Should return correct values when covariates is long enough
        ds = PastCovariatesInferenceDataset(target_series=target, covariates=long_cov,
                                            input_chunk_length=10, output_chunk_length=10, n=30)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-30])
        self.assertEqual(ds[0][3], target)

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40))
        covariate = TimeSeries.from_times_and_values(pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60))

        ds = PastCovariatesInferenceDataset(target_series=target, covariates=covariate,
                                            input_chunk_length=10, output_chunk_length=10, n=20)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:40])
        self.assertEqual(ds[0][3], target)

    def test_future_covariates_inference_dataset(self):
        # one target series
        ds = FutureCovariatesInferenceDataset(target_series=self.target1, input_chunk_length=len(self.target1))
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self.assertEqual(ds[0][1:], (None, self.target1))

        # two target series
        ds = FutureCovariatesInferenceDataset(target_series=[self.target1, self.target2],
                                              input_chunk_length=max(len(self.target1), len(self.target2)))
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self.assertEqual(ds[1][1:], (None, self.target2))

        # fail if covariates do not have same size
        with self.assertRaises(ValueError):
            ds = FutureCovariatesInferenceDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # With future past covariates:
        times1 = pd.date_range(start='20100101', end='20100701', freq='D')
        times2 = pd.date_range(start='20100101', end='20100820', freq='D')  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        short_cov = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        long_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = FutureCovariatesInferenceDataset(target_series=target, covariates=short_cov, input_chunk_length=10, n=30)

        # should fail if covariates are too short
        with self.assertRaises(ValueError):
            ds[0]

        # Should return correct values when covariates is long enough
        ds = FutureCovariatesInferenceDataset(target_series=target, covariates=long_cov, input_chunk_length=10, n=30)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-50:-20])
        self.assertEqual(ds[0][2], target)

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40))
        covariate = TimeSeries.from_times_and_values(pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60))

        ds = FutureCovariatesInferenceDataset(target_series=target, covariates=covariate, input_chunk_length=10, n=20)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[30:50])
        self.assertEqual(ds[0][2], target)

    def test_dual_covariates_inference_dataset(self):
        # one target series
        ds = DualCovariatesInferenceDataset(target_series=self.target1, input_chunk_length=len(self.target1))
        np.testing.assert_almost_equal(ds[0][0], self.vals1)
        self.assertEqual(ds[0][1:], (None, None, self.target1))

        # two target series
        ds = DualCovariatesInferenceDataset(target_series=[self.target1, self.target2],
                                            input_chunk_length=max(len(self.target1), len(self.target2)))
        np.testing.assert_almost_equal(ds[1][0], self.vals2)
        self.assertEqual(ds[1][1:], (None, None, self.target2))

        # fail if covariates do not have same size
        with self.assertRaises(ValueError):
            ds = DualCovariatesInferenceDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # With future past covariates:
        times1 = pd.date_range(start='20100101', end='20100701', freq='D')
        times2 = pd.date_range(start='20100101', end='20100820', freq='D')  # 50 days longer than times1

        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        short_cov = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        long_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = DualCovariatesInferenceDataset(target_series=target, covariates=short_cov,
                                            input_chunk_length=10, output_chunk_length=10, n=30)

        # should fail if covariates are too short
        with self.assertRaises(ValueError):
            ds[0]

        # Should return correct values when covariates is long enough
        ds = DualCovariatesInferenceDataset(target_series=target, covariates=long_cov,
                                            input_chunk_length=10, output_chunk_length=10, n=30)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], long_cov.values()[-50:-20])
        self.assertEqual(ds[0][3], target)

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40))
        covariate = TimeSeries.from_times_and_values(pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60))

        ds = DualCovariatesInferenceDataset(target_series=target, covariates=covariate,
                                            input_chunk_length=10, output_chunk_length=10, n=20)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], covariate.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], covariate.values()[30:50])
        self.assertEqual(ds[0][3], target)

    def test_mixed_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start='20100101', end='20100701', freq='D')
        times2 = pd.date_range(start='20100201', end='20100820', freq='D')  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        past_cov = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        long_past_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        future_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = MixedCovariatesInferenceDataset(target_series=target, past_covariates=past_cov,
                                             future_covariates=past_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=30)

        # should fail if future covariates are too short
        with self.assertRaises(ValueError):
            ds[0]

        # Should return correct values when covariates is long enough
        ds = MixedCovariatesInferenceDataset(target_series=target, past_covariates=long_past_cov,
                                             future_covariates=future_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=30)

        # It should contain:
        # past_target, past_covariates, historic_future_covariates, future_covariates, future_past_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][4], long_past_cov.values()[-50:-30])
        self.assertEqual(ds[0][5], target)

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40))
        past_cov = TimeSeries.from_times_and_values(pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60))
        future_cov = TimeSeries.from_times_and_values(pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70))

        ds = MixedCovariatesInferenceDataset(target_series=target, past_covariates=past_cov,
                                             future_covariates=future_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=20)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[10:20])
        np.testing.assert_almost_equal(ds[0][3], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][4], past_cov.values()[30:40])
        self.assertEqual(ds[0][5], target)

    def test_split_covariates_inference_dataset(self):
        # With future past covariates:
        times1 = pd.date_range(start='20100101', end='20100701', freq='D')
        times2 = pd.date_range(start='20100201', end='20100820', freq='D')  # ends 50 days after times1

        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        past_cov = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        long_past_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        future_cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = SplitCovariatesInferenceDataset(target_series=target, past_covariates=past_cov,
                                             future_covariates=past_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=30)

        # should fail if future covariates are too short
        with self.assertRaises(ValueError):
            ds[0]

        # Should return correct values when covariates is long enough
        ds = SplitCovariatesInferenceDataset(target_series=target, past_covariates=long_past_cov,
                                             future_covariates=future_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=30)

        # It should contain:
        # past_target, past_covariates, future_covariates, future_past_covariates
        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], long_past_cov.values()[-60:-50])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[-50:-20])
        np.testing.assert_almost_equal(ds[0][3], long_past_cov.values()[-50:-30])
        self.assertEqual(ds[0][4], target)

        # Should also work for integer-indexed series
        target = TimeSeries.from_times_and_values(pd.RangeIndex(start=10, stop=50, step=1), np.random.randn(40))
        past_cov = TimeSeries.from_times_and_values(pd.RangeIndex(start=20, stop=80, step=1), np.random.randn(60))
        future_cov = TimeSeries.from_times_and_values(pd.RangeIndex(start=30, stop=100, step=1), np.random.randn(70))

        ds = SplitCovariatesInferenceDataset(target_series=target, past_covariates=past_cov,
                                             future_covariates=future_cov, input_chunk_length=10,
                                             output_chunk_length=10, n=20)

        np.testing.assert_almost_equal(ds[0][0], target.values()[-10:])
        np.testing.assert_almost_equal(ds[0][1], past_cov.values()[20:30])
        np.testing.assert_almost_equal(ds[0][2], future_cov.values()[20:40])
        np.testing.assert_almost_equal(ds[0][3], past_cov.values()[30:40])
        self.assertEqual(ds[0][4], target)

    def test_past_covariates_sequential_dataset(self):
        # one target series
        ds = PastCovariatesSequentialDataset(target_series=self.target1, input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 81)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))

        # two target series
        ds = PastCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 262)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))
        self._assert_eq(ds[136], (self.target2[125:135], None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = PastCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             input_chunk_length=10, output_chunk_length=10, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[125:135], None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = PastCovariatesSequentialDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates
        ds = PastCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             covariates=[self.cov1, self.cov2],
                                             input_chunk_length=10, output_chunk_length=10)
        self._assert_eq(ds[5], (self.target1[75:85], self.cov1[75:85], self.target1[85:95]))
        self._assert_eq(ds[136], (self.target2[125:135], self.cov2[125:135], self.target2[135:145]))

        # should fail if covariates do not have the required time span, even though covariates are longer
        times1 = pd.date_range(start='20100101', end='20110101', freq='D')
        times2 = pd.date_range(start='20120101', end='20150101', freq='D')
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(target_series=target, covariates=cov,
                                             input_chunk_length=10, output_chunk_length=10)
        with self.assertRaises(ValueError):
            ds[5]

        # the same should fail when series are integer-indexed
        times1 = pd.RangeIndex(start=0, stop=100, step=1)
        times2 = pd.RangeIndex(start=200, stop=400, step=1)
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(target_series=target, covariates=cov,
                                             input_chunk_length=10, output_chunk_length=10)
        with self.assertRaises(ValueError):
            ds[5]

        # we should get the correct covariate slice even when target and covariates are not aligned
        times1 = pd.date_range(start='20100101', end='20110101', freq='D')
        times2 = pd.date_range(start='20090101', end='20110106', freq='D')
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(target_series=target, covariates=cov,
                                             input_chunk_length=10, output_chunk_length=10)

        self.assertTrue(all(ds[0][1] == cov.values()[-25:-15]))
        self.assertTrue(all(ds[5][1] == cov.values()[-30:-20]))

        # This should also be the case when series are integer indexed
        times1 = pd.RangeIndex(start=100, stop=200, step=1)
        times2 = pd.RangeIndex(start=50, stop=250, step=1)
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesSequentialDataset(target_series=target, covariates=cov,
                                             input_chunk_length=10, output_chunk_length=10)

        self.assertTrue(all(ds[0][1] == cov.values()[-70:-60]))
        self.assertTrue(all(ds[5][1] == cov.values()[-75:-65]))

    def test_future_covariates_sequential_dataset(self):
        # one target series
        ds = FutureCovariatesSequentialDataset(target_series=self.target1, input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 81)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))

        # two target series
        ds = FutureCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 262)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))
        self._assert_eq(ds[136], (self.target2[125:135], None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = FutureCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                               input_chunk_length=10, output_chunk_length=10, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[75:85], None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[125:135], None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = FutureCovariatesSequentialDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100))
        target2 = TimeSeries.from_values(np.random.randn(50))
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = FutureCovariatesSequentialDataset(target_series=[target1, target2], covariates=[cov1, cov2],
                                               input_chunk_length=10, output_chunk_length=10)

        self.assertTrue(all(ds[0][0] == target1.values()[-20:-10]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-30:-20]))
        self.assertTrue(all(ds[0][2] == target1.values()[-10:]))

        self.assertTrue(all(ds[101][0] == target2.values()[-40:-30]))
        self.assertTrue(all(ds[101][1] == cov2.values()[-60:-50]))
        self.assertTrue(all(ds[101][2] == target2.values()[-30:-20]))

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start='20090201', end='20090220', freq='D')
        times2 = pd.date_range(start='20090201', end='20090222', freq='D')
        target1 = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = FutureCovariatesSequentialDataset(target_series=[target1], covariates=[cov1],
                                               input_chunk_length=2, output_chunk_length=2)

        self.assertTrue(all(ds[0][0] == target1.values()[-4:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-4:-2]))
        self.assertTrue(all(ds[0][2] == target1.values()[-2:]))

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = FutureCovariatesSequentialDataset(target_series=[target1], covariates=[cov1],
                                               input_chunk_length=2, output_chunk_length=2)

        with self.assertRaises(ValueError):
            ds[0]

    def test_dual_covariates_sequential_dataset(self):
        # Must contain (past_target, historic_future_covariates, future_covariates, future_target)

        # one target series
        ds = DualCovariatesSequentialDataset(target_series=self.target1, input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 81)
        self._assert_eq(ds[5], (self.target1[75:85], None, None, self.target1[85:95]))

        # two target series
        ds = DualCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             input_chunk_length=10, output_chunk_length=10)
        self.assertEqual(len(ds), 262)
        self._assert_eq(ds[5], (self.target1[75:85], None, None, self.target1[85:95]))
        self._assert_eq(ds[136], (self.target2[125:135], None, None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = DualCovariatesSequentialDataset(target_series=[self.target1, self.target2],
                                             input_chunk_length=10, output_chunk_length=10, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[75:85], None, None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[125:135], None, None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = DualCovariatesSequentialDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates; covariates not aligned, must contain correct values
        target1 = TimeSeries.from_values(np.random.randn(100))
        target2 = TimeSeries.from_values(np.random.randn(50))
        cov1 = TimeSeries.from_values(np.random.randn(120))
        cov2 = TimeSeries.from_values(np.random.randn(80))

        ds = DualCovariatesSequentialDataset(target_series=[target1, target2], covariates=[cov1, cov2],
                                             input_chunk_length=10, output_chunk_length=10)

        self.assertTrue(all(ds[0][0] == target1.values()[-20:-10]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-40:-30]))
        self.assertTrue(all(ds[0][2] == cov1.values()[-30:-20]))
        self.assertTrue(all(ds[0][3] == target1.values()[-10:]))

        self.assertTrue(all(ds[101][0] == target2.values()[-40:-30]))
        self.assertTrue(all(ds[101][1] == cov2.values()[-70:-60]))
        self.assertTrue(all(ds[101][2] == cov2.values()[-60:-50]))
        self.assertTrue(all(ds[101][3] == target2.values()[-30:-20]))

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start='20090201', end='20090220', freq='D')
        times2 = pd.date_range(start='20090201', end='20090222', freq='D')
        target1 = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))

        ds = DualCovariatesSequentialDataset(target_series=[target1], covariates=[cov1],
                                             input_chunk_length=2, output_chunk_length=2)

        self.assertTrue(all(ds[0][0] == target1.values()[-4:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-6:-4]))
        self.assertTrue(all(ds[0][2] == cov1.values()[-4:-2]))
        self.assertTrue(all(ds[0][3] == target1.values()[-2:]))

        # Should fail if covariates are not long enough
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(7))

        ds = DualCovariatesSequentialDataset(target_series=[target1], covariates=[cov1],
                                             input_chunk_length=2, output_chunk_length=2)

        with self.assertRaises(ValueError):
            ds[0]

    def test_past_covariates_shifted_dataset(self):
        # one target series
        ds = PastCovariatesShiftedDataset(target_series=self.target1, length=10, shift=5)
        self.assertEqual(len(ds), 86)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))

        # two target series
        ds = PastCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10, shift=5)
        self.assertEqual(len(ds), 272)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = PastCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10,
                                          shift=5, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[130:140], None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = PastCovariatesShiftedDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates
        ds = PastCovariatesShiftedDataset(target_series=[self.target1, self.target2],
                                          covariates=[self.cov1, self.cov2],
                                          length=10, shift=5)
        self._assert_eq(ds[5], (self.target1[80:90], self.cov1[80:90], self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], self.cov2[130:140], self.target2[135:145]))

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = PastCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-7:-4]))
        self.assertTrue(all(ds[0][2] == target1.values()[-3:]))

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start='20090201', end='20090220', freq='D')
        times2 = pd.date_range(start='20090201', end='20090222', freq='D')
        target1 = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = PastCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-7:-4]))
        self.assertTrue(all(ds[0][2] == target1.values()[-3:]))

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(5))
        ds = PastCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        with self.assertRaises(ValueError):
            ds[0]

    def test_future_covariates_shifted_dataset(self):
        # one target series
        ds = FutureCovariatesShiftedDataset(target_series=self.target1, length=10, shift=5)
        self.assertEqual(len(ds), 86)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))

        # two target series
        ds = FutureCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10, shift=5)
        self.assertEqual(len(ds), 272)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = FutureCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10,
                                            shift=5, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[80:90], None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[130:140], None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = FutureCovariatesShiftedDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates
        ds = FutureCovariatesShiftedDataset(target_series=[self.target1, self.target2],
                                            covariates=[self.cov1, self.cov2],
                                            length=10, shift=5)
        self._assert_eq(ds[5], (self.target1[80:90], self.cov1[85:95], self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], self.cov2[135:145], self.target2[135:145]))

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = FutureCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                            length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-5:-2]))
        self.assertTrue(all(ds[0][2] == target1.values()[-3:]))

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start='20090201', end='20090220', freq='D')
        times2 = pd.date_range(start='20090201', end='20090222', freq='D')
        target1 = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = FutureCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                            length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-5:-2]))
        self.assertTrue(all(ds[0][2] == target1.values()[-3:]))

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = FutureCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                            length=3, shift=2)
        with self.assertRaises(ValueError):
            ds[0]

    def test_dual_covariates_shifted_dataset(self):
        # one target series
        ds = DualCovariatesShiftedDataset(target_series=self.target1, length=10, shift=5)
        self.assertEqual(len(ds), 86)
        self._assert_eq(ds[5], (self.target1[80:90], None, None, self.target1[85:95]))

        # two target series
        ds = DualCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10, shift=5)
        self.assertEqual(len(ds), 272)
        self._assert_eq(ds[5], (self.target1[80:90], None, None, self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], None, None, self.target2[135:145]))

        # two target series with custom max_nr_samples
        ds = DualCovariatesShiftedDataset(target_series=[self.target1, self.target2], length=10,
                                          shift=5, max_samples_per_ts=50)
        self.assertEqual(len(ds), 100)
        self._assert_eq(ds[5], (self.target1[80:90], None, None, self.target1[85:95]))
        self._assert_eq(ds[55], (self.target2[130:140], None, None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = DualCovariatesShiftedDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates
        ds = DualCovariatesShiftedDataset(target_series=[self.target1, self.target2],
                                          covariates=[self.cov1, self.cov2],
                                          length=10, shift=5)
        self._assert_eq(ds[5], (self.target1[80:90], self.cov1[80:90], self.cov1[85:95], self.target1[85:95]))
        self._assert_eq(ds[141], (self.target2[130:140], self.cov2[130:140], self.cov2[135:145], self.target2[135:145]))

        # Should contain correct values even when covariates are not aligned
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(10))
        ds = DualCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-7:-4]))
        self.assertTrue(all(ds[0][2] == cov1.values()[-5:-2]))
        self.assertTrue(all(ds[0][3] == target1.values()[-3:]))

        # Should also contain correct values when time-indexed with covariates not aligned
        times1 = pd.date_range(start='20090201', end='20090220', freq='D')
        times2 = pd.date_range(start='20090201', end='20090222', freq='D')
        target1 = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov1 = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        ds = DualCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        self.assertTrue(all(ds[0][0] == target1.values()[-5:-2]))
        self.assertTrue(all(ds[0][1] == cov1.values()[-7:-4]))
        self.assertTrue(all(ds[0][2] == cov1.values()[-5:-2]))
        self.assertTrue(all(ds[0][3] == target1.values()[-3:]))

        # Should fail if covariates are too short
        target1 = TimeSeries.from_values(np.random.randn(8))
        cov1 = TimeSeries.from_values(np.random.randn(7))
        ds = DualCovariatesShiftedDataset(target_series=[target1], covariates=[cov1],
                                          length=3, shift=2)
        with self.assertRaises(ValueError):
            ds[0]

    def test_horizon_based_dataset(self):
        # one target series
        ds = HorizonBasedDataset(target_series=self.target1, output_chunk_length=10, lh=(1, 3), lookback=2)
        self.assertEqual(len(ds), 20)
        self._assert_eq(ds[5], (self.target1[65:85], None, self.target1[85:95]))

        # two target series
        ds = HorizonBasedDataset(target_series=[self.target1, self.target2],
                                 output_chunk_length=10, lh=(1, 3), lookback=2)
        self.assertEqual(len(ds), 40)
        self._assert_eq(ds[5], (self.target1[65:85], None, self.target1[85:95]))
        self._assert_eq(ds[25], (self.target2[115:135], None, self.target2[135:145]))

        # two targets and one covariate
        with self.assertRaises(ValueError):
            ds = HorizonBasedDataset(target_series=[self.target1, self.target2], covariates=[self.cov1])

        # two targets and two covariates
        ds = HorizonBasedDataset(target_series=[self.target1, self.target2],
                                 covariates=[self.cov1, self.cov2],
                                 output_chunk_length=10, lh=(1, 3), lookback=2)
        self._assert_eq(ds[5], (self.target1[65:85], self.cov1[65:85], self.target1[85:95]))
        self._assert_eq(ds[25], (self.target2[115:135], self.cov2[115:135], self.target2[135:145]))

    def test_get_matching_index(self):
        from ..utils.data.utils import _get_matching_index

        # Check dividable freq
        times1 = pd.date_range(start='20100101', end='20100330', freq='D')
        times2 = pd.date_range(start='20100101', end='20100320', freq='D')
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        self.assertEqual(_get_matching_index(target, cov, idx=15), 5)

        # check non-dividable freq
        times1 = pd.date_range(start='20100101', end='20120101', freq='M')
        times2 = pd.date_range(start='20090101', end='20110601', freq='M')
        target = TimeSeries.from_times_and_values(times1, np.random.randn(len(times1)))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        self.assertEqual(_get_matching_index(target, cov, idx=15), 15 - 7)

        # check integer-indexed series
        times2 = pd.RangeIndex(start=10, stop=90)
        target = TimeSeries.from_values(np.random.randn(100))
        cov = TimeSeries.from_times_and_values(times2, np.random.randn(len(times2)))
        self.assertEqual(_get_matching_index(target, cov, idx=15), 5)



import shutil

import numpy as np
import pandas as pd

from .base_test_class import DartsBaseTestClass
from ..timeseries import TimeSeries
from ..utils import timeseries_generation as tg
from ..metrics import mape
from ..models import (
    NaiveSeasonal,
    ExponentialSmoothing,
    ARIMA,
    Theta,
    FourTheta,
    FFT,
    VARIMA,
    RandomForest,
    LinearRegressionModel,
)
from ..utils.utils import SeasonalityMode, TrendMode, ModelMode
from ..logging import get_logger
from ..datasets import AirPassengersDataset, IceCreamHeaterDataset

logger = get_logger(__name__)

# (forecasting models, maximum error) tuples
models = [
    (ExponentialSmoothing(), 5.6),
    (ARIMA(12, 2, 1), 10),
    (ARIMA(1, 1, 1), 40),
    (Theta(), 11.3),
    (Theta(1), 20.2),
    (Theta(-1), 9.8),
    (FourTheta(1), 20.2),
    (FourTheta(-1), 9.8),
    (FourTheta(trend_mode=TrendMode.EXPONENTIAL), 5.5),
    (FourTheta(model_mode=ModelMode.MULTIPLICATIVE), 11.4),
    (FourTheta(season_mode=SeasonalityMode.ADDITIVE), 14.2),
    (FFT(trend="poly"), 11.4),
    (NaiveSeasonal(), 32.4),
    (LinearRegressionModel(lags=12), 11.0),
    (RandomForest(lags=12, n_estimators=200, max_depth=3), 15.5),
]
# forecasting models with exogenous variables support
multivariate_models = [
    (VARIMA(1, 0, 0), 55.6),
    (VARIMA(1, 1, 1), 57.0),
]

dual_models = [ARIMA()]


try:
    from ..models import Prophet
    models.append((Prophet(), 13.5))
except ImportError:
    logger.warning("Prophet not installed - will be skipping Prophet tests")

try:
    from ..models import AutoARIMA

    models.append((AutoARIMA(), 12.2))
    dual_models.append(AutoARIMA())
    PMDARIMA_AVAILABLE = True
except ImportError:
    logger.warning("pmdarima not installed - will be skipping AutoARIMA tests")
    PMDARIMA_AVAILABLE = False

try:
    from ..models import TCNModel

    TORCH_AVAILABLE = True
except ImportError:
    logger.warning("Torch not installed - will be skipping Torch models tests")
    TORCH_AVAILABLE = False


class LocalForecastingModelsTestCase(DartsBaseTestClass):

    # forecasting horizon used in runnability tests
    forecasting_horizon = 5

    # dummy timeseries for runnability tests
    np.random.seed(1)
    ts_gaussian = tg.gaussian_timeseries(length=100, mean=50)

    # real timeseries for functionality tests
    ts_passengers = AirPassengersDataset().load()
    ts_pass_train, ts_pass_val = ts_passengers.split_after(pd.Timestamp("19570101"))

    # real multivariate timeseries for functionality tests
    ts_ice_heater = IceCreamHeaterDataset().load()
    ts_ice_heater_train, ts_ice_heater_val = ts_ice_heater.split_after(split_point=0.7)

    def test_models_runnability(self):
        for model, _ in models:
            model.fit(self.ts_gaussian)
            prediction = model.predict(self.forecasting_horizon)
            self.assertTrue(len(prediction) == self.forecasting_horizon)

    def test_models_performance(self):
        # for every model, check whether its errors do not exceed the given bounds
        for model, max_mape in models:
            np.random.seed(1)  # some models are probabilist...
            model.fit(self.ts_pass_train)
            prediction = model.predict(len(self.ts_pass_val))
            current_mape = mape(prediction, self.ts_pass_val)
            self.assertTrue(current_mape < max_mape, "{} model exceeded the maximum MAPE of {}. "
                            "with a MAPE of {}".format(str(model), max_mape, current_mape))

    def test_multivariate_models_performance(self):
        # for every model, check whether its errors do not exceed the given bounds
        for model, max_mape in multivariate_models:
            np.random.seed(1)
            model.fit(self.ts_ice_heater_train)
            prediction = model.predict(len(self.ts_ice_heater_val))
            current_mape = mape(prediction, self.ts_ice_heater_val)
            self.assertTrue(current_mape < max_mape, "{} model exceeded the maximum MAPE of {}. "
                            "with a MAPE of {}".format(str(model), max_mape, current_mape))

    def test_multivariate_input(self):
        es_model = ExponentialSmoothing()
        ts_passengers_enhanced = self.ts_passengers.add_datetime_attribute("month")
        with self.assertRaises(AssertionError):
            es_model.fit(ts_passengers_enhanced)
        es_model.fit(ts_passengers_enhanced["#Passengers"])
        with self.assertRaises(KeyError):
            es_model.fit(ts_passengers_enhanced["2"])

    def test_exogenous_variables_support(self):
        for model in dual_models:

            # Test models runnability
            model.fit(self.ts_gaussian, future_covariates=self.ts_gaussian)

            prediction = model.predict(
                self.forecasting_horizon,
                future_covariates=tg.gaussian_timeseries(
                    length=self.forecasting_horizon,
                    start=self.ts_gaussian.end_time() + self.ts_gaussian.freq))

            self.assertTrue(len(prediction) == self.forecasting_horizon)

            # Test mismatch in length between exogenous variables and forecasting horizon
            with self.assertRaises(ValueError):
                model.predict(
                    self.forecasting_horizon,
                    future_covariates=tg.gaussian_timeseries(length=self.forecasting_horizon - 1))

            # Test mismatch in time-index/length between series and exogenous variables
            with self.assertRaises(ValueError):
                model.fit(self.ts_gaussian, future_covariates=self.ts_gaussian[:-1])
            with self.assertRaises(ValueError):
                model.fit(self.ts_gaussian[1:], future_covariates=self.ts_gaussian[:-1])

    def test_dummy_series(self):
        values = np.random.uniform(low=-10, high=10, size=100)
        ts = TimeSeries.from_dataframe(pd.DataFrame({"V1": values}))

        varima = VARIMA(trend="t")
        with self.assertRaises(ValueError):
            varima.fit(series=ts)

        if PMDARIMA_AVAILABLE:
            autoarima = AutoARIMA(trend="t")
            with self.assertRaises(ValueError):
                autoarima.fit(series=ts)



import unittest
import pandas as pd

from darts.dataprocessing.transformers.mappers import Mapper, InvertibleMapper
from darts.utils.timeseries_generation import constant_timeseries, linear_timeseries


class MappersTestCase(unittest.TestCase):

    @staticmethod
    def func(x):
        return x + 10

    @staticmethod
    def inverse_func(x):
        return x - 10

    @staticmethod
    def ts_func(ts, x):
        return x - ts.month

    @staticmethod
    def inverse_ts_func(ts, x):
        return x + ts.month

    plus_ten = Mapper(func.__func__)
    plus_ten_invertible = InvertibleMapper(func.__func__, inverse_func.__func__)

    subtract_month = Mapper(ts_func.__func__)
    subtract_month_invertible = InvertibleMapper(ts_func.__func__, inverse_ts_func.__func__)

    lin_series = linear_timeseries(start_value=1, length=12, freq='MS', start=pd.Timestamp('2000-01-01'), end_value=12)  # noqa: E501
    zeroes = constant_timeseries(value=0.0, length=12, freq='MS', start=pd.Timestamp('2000-01-01'))
    tens = constant_timeseries(value=10.0, length=12, freq='MS', start=pd.Timestamp('2000-01-01'))
    twenties = constant_timeseries(value=20.0, length=12, freq='MS', start=pd.Timestamp('2000-01-01'))

    def test_mapper(self):

        test_cases = [
            (self.zeroes, self.tens),
            ([self.zeroes, self.tens], [self.tens, self.twenties])
        ]

        for to_transform, expected_output in test_cases:
            transformed = self.plus_ten.transform(to_transform)
            self.assertEqual(transformed, expected_output)

    def test_invertible_mapper(self):
        test_cases = [
            (self.zeroes),
            ([self.zeroes, self.tens])
        ]

        for data in test_cases:
            transformed = self.plus_ten_invertible.transform(data)
            back = self.plus_ten_invertible.inverse_transform(transformed)
            self.assertEqual(back, data)

    def test_mapper_with_timestamp(self):

        test_cases = [
            (self.lin_series, self.zeroes),
            ([self.lin_series, self.lin_series], [self.zeroes, self.zeroes])
        ]

        for to_transform, expected_output in test_cases:
            transformed = self.subtract_month.transform(to_transform)
            if isinstance(to_transform, list):
                expected_output = [o.with_columns_renamed(o.components[0], t.components[0]) for t, o in zip(transformed, expected_output)]
            else:
                expected_output = expected_output.with_columns_renamed(expected_output.components[0], transformed.components[0])
            self.assertEqual(transformed, expected_output)

    def test_invertible_mapper_with_timestamp(self):

        test_cases = [
            (self.lin_series),
            ([self.lin_series, self.lin_series])
        ]

        for data in test_cases:
            transformed = self.subtract_month_invertible.transform(data)
            back = self.subtract_month_invertible.inverse_transform(transformed)
            self.assertEqual(back, data)
